from typing import Self

from PyQt6.QtCore import QSize
from PyQt6.QtWidgets import QFrame, QLayout, QScrollArea, QTabWidget, QVBoxLayout, QWidget

from .augmented import Widget
from .markdown_view import MarkdownView


class ParameterDescriptionWidget(Widget):
    """
    Widget with two tabs: one for parameters and one for description text.

    Parameters
    ----------
    parameter_layout
        The layout to use for the parameter tab.
    parameter_tab_name
        The text used for the parameter tab name.
    """

    def __init__(self, parameter_layout: QLayout | None, parameter_tab_name: str = "Parameters"):
        layout = QVBoxLayout()
        Widget.__init__(self, layout)
        self.param_widget = Widget(parameter_layout)
        parameter_scroll_area = QScrollArea()
        parameter_scroll_area.setWidget(self.param_widget)
        parameter_scroll_area.setWidgetResizable(True)
        parameter_scroll_area.setFrameShape(QFrame.Shape.NoFrame)

        self.description_widget = MarkdownView()

        tab_widget = QTabWidget()
        layout.addWidget(tab_widget)
        tab_widget.addTab(parameter_scroll_area, parameter_tab_name)
        tab_widget.addTab(self.description_widget, "Description")

    def parameter_widget(self) -> QWidget:
        """Get the parameter widget (the first tab)."""
        return self.param_widget

    def set_description(self, text: str) -> Self:
        """Set the text (interpreted as Markdown) displayed in the description tab."""
        self.description_widget.setMarkdown(text)
        return self

    def sizeHint(self) -> QSize:
        return Widget.sizeHint(self) * 1.5



from __future__ import annotations

import asyncio
import time
from abc import abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol

if TYPE_CHECKING:
    from .step_runner import StepRunner


class SequenceStep(Protocol):
    """Represents a sequence step that the sequence can run."""

    @abstractmethod
    async def run(self, runner: StepRunner, data_directory: Path):
        """
        Run the sequence step. Ensure you call `self.sleep()` often enough that other tasks can be
        run.
        """
        ...

    @abstractmethod
    def reset(self):
        """
        Reset this step to its original state (i.e. before `run()` was called). Sequence steps might
        be re-used, so they must be resettable. An easy implementation is to call `__init__()` to
        reinitialize with the original parameters.
        """
        ...

    @abstractmethod
    def name(self) -> str:
        """A name used to represent this step. For example, "Hold" or "Increment Temperature"."""
        ...

    def directory_name(self) -> str:
        """
        Get the name (i.e. not full path) of the directory where data should be recorded. By default
        this just returns the result of `name()`.
        """
        return self.name()

    def metadata(self) -> dict[str, Any]:
        """Get metadata for this sequence step to be recorded in a JSON file."""
        return {}

    async def sleep(self, delay: float):
        """
        Sleep this sequence step for **delay** seconds. This is *not* equivalent to `time.sleep()`
        and should be used *instead* of `time.sleep()`.

        Do not override this.
        """
        await asyncio.sleep(delay)

    async def sleep_until(self, when: float):
        """
        Sleep until **when**, which is a `float` as returned by `time.time()`.

        Do not override this.
        """
        await self.sleep(0)  # make sure we sleep at all
        await self.sleep(when - time.time())



from typing import Any, Callable

from PyQt6.QtCore import QSize
from PyQt6.QtWidgets import QPushButton, QSizePolicy


class Button(QPushButton):
    """
    `QPushButton` that automatically connects any provided functions to the `clicked` signal.

    Parameters
    ----------
    text
        The text to display on the button.
    push_fn
        Function(s) to connect the `clicked` signal to.
    """

    def __init__(self, text: str, *push_fn: Callable[[], None | Any]):
        QPushButton.__init__(self, text)
        for fn in push_fn:
            self.clicked.connect(fn)


class FixedButton(Button):
    """Button with a fixed size."""

    def __init__(self, text: str, *push_fn: Callable[[], None | Any]):
        Button.__init__(self, text, *push_fn)
        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)


class BiggerButton(Button):  # this name feels so dumb xD
    """Button with user-selected size scaling."""

    def __init__(
        self,
        text: str,
        *push_fn: Callable[[], None | Any],
        size_scalars: tuple[float, float] = (1, 1),
    ):
        Button.__init__(self, text, *push_fn)
        self.horizontal_scalar, self.vertical_scalar = size_scalars

    def set_size_scalars(self, size_scalars: tuple[float, float]):
        """
        Set the vertical and horizontal size scalars.

        Parameters
        ----------
        size_scalars
            The new vertical and horizontal (respectively) size scalars.
        """
        self.horizontal_scalar, self.vertical_scalar = size_scalars

    def sizeHint(self) -> QSize:  # overridden
        default_size = Button.sizeHint(self)
        return QSize(
            round(default_size.width() * self.horizontal_scalar),
            round(default_size.height() * self.vertical_scalar),
        )



import sys

from PyQt6.QtWidgets import QAbstractSpinBox, QDoubleSpinBox, QPushButton, QSpinBox


class DoubleSpinBox(QDoubleSpinBox):
    """
    `QDoubleSpinBox` without up/down buttons.

    Parameters
    ----------
    number_of_decimals
        The number of decimals to display.
    minimum
        The minimum allowed value (default 0).
    maximum
        The maximum allowed value (default largest possible float).
    initial_value
        The initial value of the spinbox (default higher of 0 and **minimum**).
    """

    LARGEST_FLOAT = sys.float_info.max

    def __init__(
        self,
        number_of_decimals: int = 1,
        minimum: float = 0,
        maximum: float = LARGEST_FLOAT,
        initial_value: int = 0,
    ):
        QDoubleSpinBox.__init__(self)
        self.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.setDecimals(number_of_decimals)
        self.setMinimum(minimum)
        self.setMaximum(maximum)
        self.setValue(initial_value)

    def connect_to_button(self, button: QPushButton):
        """
        Emit the **button**'s `pressed` signal when Enter is pressed and the button is enabled.
        """
        line_edit = self.lineEdit()
        if line_edit is not None:
            line_edit.returnPressed.connect(
                lambda: button.pressed.emit() if button.isEnabled() else None
            )


class SpinBox(QSpinBox):
    """
    `QSpinBox` class without up/down buttons.

    Parameters
    ----------
    minimum
        The minimum allowed value (default 0).
    maximum
        The maximum allowed value (default largest possible integer).
    initial_value
        The initial value of the spinbox (default higher of 0 and **minimum**).
    """

    LARGEST_INTEGER = 2147483647

    def __init__(self, minimum: int = 0, maximum: int = LARGEST_INTEGER, initial_value: int = 0):
        QSpinBox.__init__(self)
        self.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
        self.setMinimum(minimum)
        self.setMaximum(maximum)
        self.setValue(initial_value)



"""Headers for metadata."""

from .datatime import HEADER

START_DATETIME = f"Start Time ({HEADER})"
END_DATETIME = f"End Time ({HEADER})"
DURATION = "Duration (seconds)"
SETPOINT = "Oven Setpoint (degrees C)"



"""Filepaths and dictionary keys for Gamry settings."""

from .defaults import DEFAULT_SETTINGS
from .saved import SAVED_SETTINGS

SETTINGS_FOLDER_NAME = "gamry"
SETTINGS_FILENAME = "gamry_settings.json"

# files
SAVED_SETTINGS_FOLDER = SAVED_SETTINGS.joinpath(SETTINGS_FOLDER_NAME)
SAVED_SETTINGS_FILE = SAVED_SETTINGS.joinpath(SETTINGS_FILENAME)
DEFAULT_SETTINGS_FILE = DEFAULT_SETTINGS.joinpath(SETTINGS_FOLDER_NAME, SETTINGS_FILENAME)
DESCRIPTION_FOLDER = DEFAULT_SETTINGS.joinpath(SETTINGS_FOLDER_NAME)
DESCRIPTION_FILENAME = "description.md"
# keys
ENABLED = "Enabled"
LOCATION = "GamryCOM Location"
# other
DEFAULT_GAMRY_LOCATION = "C:/Program Files (x86)/Gamry Instruments/Framework/GamryCom.exe"



from typing import TYPE_CHECKING

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QMenuBar

from ..classes import Action
from .file import FileMenu
from .sequence import SequenceMenu
from .view import ViewMenu

if TYPE_CHECKING:
    from ..main_window import MainWindow


class MenuBar(QMenuBar):
    """The application's menubar."""

    def __init__(self, main_window: "MainWindow"):
        QMenuBar.__init__(self)
        self.create_submenus(main_window)

    def create_submenus(self, main_window: "MainWindow"):
        # need to store these because they have signals
        self.sequence = SequenceMenu(self)
        self.view = ViewMenu(self, main_window)
        self.addMenu(FileMenu(self, main_window))
        self.addMenu(self.view)
        self.addMenu(self.sequence)

        self.right_menu_bar = QMenuBar(self)
        self.setCornerWidget(self.right_menu_bar, Qt.Corner.TopRightCorner)
        self.right_menu_bar.addAction(Action(self, "Settings", main_window.show_settings))



from __future__ import annotations

from dataclasses import dataclass
from typing import Self

from PyQt6.QtCore import QItemSelection, QModelIndex, QPersistentModelIndex, QPoint, Qt
from PyQt6.QtGui import QDragMoveEvent, QKeyEvent
from PyQt6.QtWidgets import QFileDialog, QHBoxLayout, QVBoxLayout

from ...classes import Shortcut, Timer
from ...constants.paths.settings import sequence as sequence_paths
from ...custom_widgets import Button, Container, FixedButton, OkDialog
from ...utility import images, layout as layout_util
from ..tree_models import SequenceModel
from .tree_view import TreeView


@dataclass
class DragTracker:
    timer: Timer
    position: QPoint


class SequenceTreeView(TreeView[SequenceModel]):
    """
    Custom `TreeView` for displaying sequence settings.

    Parameters
    ----------
    model
        The model the view will use.
    """

    def __init__(self, model: SequenceModel):
        TreeView.__init__(self, model)
        # configure
        self.setAcceptDrops(True)
        self.setDefaultDropAction(Qt.DropAction.MoveAction)

        # used to expand items when hovering over them
        point = QPoint()
        timer = Timer(self, 500, lambda: self.expand(self.indexAt(point)))
        timer.setSingleShot(True)
        self.drag_tracker = DragTracker(timer, point)

    @classmethod
    def from_autosave(cls) -> Self:
        """Create from the autosave files."""
        model = SequenceModel([])
        model.init_from_json(sequence_paths.SEQUENCE_ITEMS_FILE)
        view = cls(model)
        view.init_view_state_from_json(sequence_paths.SEQUENCE_STATE_FILE)
        return view

    def select_load(self):
        """Load sequence items from a user-selected file."""
        file, _ = QFileDialog.getOpenFileName(
            self, "Select sequence file", filter="JSON files (*.json)"
        )
        if file != "":
            if not self.model().init_from_json(file):
                OkDialog(
                    "Load Error",
                    "Failed to load the requested file. Ensure the file is correctly formatted and "
                    "that all plugins the file references are installed.",
                ).exec()
            else:
                self.expandAll()

    def select_save(self):
        """Save the sequence items to a user-selected file."""
        file, _ = QFileDialog.getSaveFileName(
            self, "Select save file", "untitled.json", "JSON files (*.json)"
        )
        if file != "":
            if not self.model().to_json(file):
                OkDialog(
                    "Save Error",
                    "Failed to save sequence. "
                    "This could be caused by a faulty plugin or permission issues. "
                    "See the error log for details.",
                ).exec()

    def save_on_close(self):
        """Call this when closing the application to save the item and view state."""
        self.model().to_json(sequence_paths.SEQUENCE_ITEMS_FILE)
        self.save_view_state_to_json(sequence_paths.SEQUENCE_STATE_FILE)

    def connect_signals(self):  # overridden
        TreeView.connect_signals(self)
        self.model().itemAdded.connect(self.handle_new_item)

    def create_shortcuts(self):  # overridden
        TreeView.create_shortcuts(self)
        Shortcut(
            self, "Ctrl+X", self.cut_event, context=Qt.ShortcutContext.WidgetWithChildrenShortcut
        )
        Shortcut(
            self, "Ctrl+V", self.paste_event, context=Qt.ShortcutContext.WidgetWithChildrenShortcut
        )

    def items_editable(self) -> bool:  # overridden
        return True  # TODO: make this depend on whether the widget is enabled

    def handle_new_item(self, index: QModelIndex):
        """Handle an item being added to the model."""
        self.expand(index)  # expand the new item
        self.expand(index.parent())  # and its parent

    def cut_event(self):
        """Move items to the clipboard."""
        self.copy_event()
        self.delete_event()

    def paste_event(self):
        """Paste items from the clipboard after the currently selected item."""
        self.model().paste_items(self.currentIndex())

    def delete_event(self):
        """Delete currently selected items."""
        # store the index below the current index
        persistent_new_selection_index = QPersistentModelIndex(self.indexBelow(self.currentIndex()))

        self.model().delete_items(self.selectedIndexes())

        # select the next available item after deleting
        new_selection_index = QModelIndex(persistent_new_selection_index)
        if not new_selection_index.isValid():
            # try the item below the currently selected item
            new_selection_index = self.indexBelow(self.currentIndex())
            if not new_selection_index.isValid():
                # try whatever is currently selected (usually the last item in this situation)
                new_selection_index = self.currentIndex()
                if not new_selection_index.isValid():
                    # at this point there should be no items in the model
                    self.clearSelection()

        self.setCurrentIndex(new_selection_index)

    # ----------------------------------------------------------------------------------------------
    def keyPressEvent(self, event: QKeyEvent | None):  # overridden
        if event is not None:
            match event.key():
                case Qt.Key.Key_Delete:  # delete the current item
                    self.delete_event()
                case Qt.Key.Key_Return | Qt.Key.Key_Enter:
                    self.open_event(self.selectedIndexes())
        TreeView.keyPressEvent(self, event)

    def dragMoveEvent(self, event: QDragMoveEvent | None):  # overridden
        if event is not None:
            tracked_position = self.drag_tracker.position
            event_position = event.position().toPoint()
            tracked_position.setX(event_position.x())
            tracked_position.setY(event_position.y())
            self.drag_tracker.timer.start()
        TreeView.dragMoveEvent(self, event)


class SequenceTreeWidget(Container):
    """SequenceTreeView with a delete button."""

    def __init__(self, view: SequenceTreeView):
        layout = QVBoxLayout()
        Container.__init__(self, layout)

        self.view = view
        self.delete_button = Button("Delete Selected Items", self.view.delete_event)
        self.create_widgets(layout)
        self.connect_signals()

    @classmethod
    def from_autosave(cls) -> Self:
        """Create from the autosave files."""
        view = SequenceTreeView.from_autosave()
        return cls(view)

    def create_widgets(self, layout: QVBoxLayout):
        """Create widgets at construction."""
        button_layout = QHBoxLayout()
        button_container = Container(button_layout)

        self.delete_button.setEnabled(False)
        button_layout.addWidget(self.delete_button)

        button_sublayout = QHBoxLayout()
        self.save_button = FixedButton("Save", self.view.select_save)
        self.save_button.setIcon(images.make_icon("script-export.png"))
        button_sublayout.addWidget(self.save_button)
        self.load_button = FixedButton("Load", self.view.select_load)
        self.load_button.setIcon(images.make_icon("script-import.png"))
        button_sublayout.addWidget(self.load_button)

        button_layout.addWidget(self.delete_button, alignment=Qt.AlignmentFlag.AlignLeft)
        button_layout.addLayout(button_sublayout)

        layout_util.add_to_layout(layout, button_container, self.view)

    def connect_signals(self):
        """Connect signals at construction."""
        self.view.selectionModel().selectionChanged.connect(  # type: ignore
            self.handle_selection_change
        )

    def handle_selection_change(self, selected: QItemSelection, *args):
        """Handle the item selection changing."""
        enabled = True
        if selected.isEmpty():
            enabled = False
        else:
            for index in selected.indexes():
                if not index.isValid():
                    enabled = False
        self.delete_button.setEnabled(enabled)

    def save_on_close(self):
        """Call this when closing the application to save settings."""
        self.view.save_on_close()



from PyQt6.QtWidgets import QFormLayout, QVBoxLayout

from ..custom_widgets import GroupBox, Label, StabilityProgressBar
from ..enums import StabilityStatus
from ..instruments import INSTRUMENTS
from ..utility import layout as layout_util


class StabilityCheckWidget(GroupBox):
    """Widget for monitoring whether the oven's temperature is stable."""

    NULL_TEXT = "----------"

    def __init__(self) -> None:
        GroupBox.__init__(self, "Oven Stability", QVBoxLayout())
        self.previous_count = 0
        self.status_label: Label
        self.progress_bar: StabilityProgressBar
        self.create_widgets()
        self.connect_signals()

    def create_widgets(self):
        """Create subwidgets."""
        layout: QVBoxLayout = self.layout()  # type: ignore
        label_layout = layout_util.add_sublayout(layout, QFormLayout())
        self.status_label = Label(self.NULL_TEXT)
        self.progress_bar = StabilityProgressBar(INSTRUMENTS.oven)
        self.handle_stability_change(False)  # initialize text and color

        label_layout.addRow("Status:", self.status_label)
        layout.addWidget(self.progress_bar)

    def connect_signals(self):
        """Connect widget signals."""
        oven = INSTRUMENTS.oven
        oven.stabilityChanged.connect(self.handle_stability_change)
        oven.stabilityCountChanged.connect(self.handle_stability_count_change)
        oven.connectionChanged.connect(self.handle_connection_change)

    def handle_stability_change(self, stable: bool):
        """Handle a stability change from the oven."""
        if INSTRUMENTS.oven.is_connected():
            text = StabilityStatus.bool_to_str(stable)
            color = StabilityStatus.bool_to_color(stable)
            self.status_label.setText(text)
            self.status_label.set_color(color)

    def handle_stability_count_change(self, count: int):
        """Handle the oven's stability count changing."""
        self.progress_bar.setValue(count)

    def handle_connection_change(self, connected: bool):
        """Handle the oven disconnecting/reconnecting."""
        if not connected:
            self.status_label.reset(self.NULL_TEXT)
        else:
            self.handle_stability_change(INSTRUMENTS.oven.is_stable())
            self.handle_stability_count_change(INSTRUMENTS.oven.stability_count())



import json
import os
import sys

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QCheckBox, QFileDialog, QHBoxLayout, QVBoxLayout

from ...constants import APP_NAME
from ...constants.paths.settings import gamry
from ...gamry_integration import GAMRY
from ...utility import images, layout as layout_util
from ..augmented import Button, IconLabel, OkCancelDialog
from .settings_description import SettingsDescriptionWidget


class GamrySettingsTab(SettingsDescriptionWidget):
    """Gamry-related settings."""

    def __init__(self):
        try:
            with open(gamry.SAVED_SETTINGS_FILE, "r") as f:
                settings_dict = json.load(f)
        except Exception:
            with open(gamry.DEFAULT_SETTINGS_FILE, "r") as f:
                settings_dict = json.load(f)

        layout = QVBoxLayout()
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        SettingsDescriptionWidget.__init__(self, layout)

        self.set_description_from_file(
            gamry.DESCRIPTION_FOLDER,
            gamry.DESCRIPTION_FILENAME,
            {"APPLICATION_NAME": APP_NAME},
        )

        self.enabled_checkbox = QCheckBox("Enable Gamry features")
        self.enabled_checkbox.setChecked(settings_dict[gamry.ENABLED])
        layout.addWidget(self.enabled_checkbox)

        self.gamry_location_label = IconLabel(
            images.make_pixmap("document-code.png"), settings_dict[gamry.LOCATION]
        )
        button_layout = layout_util.add_sublayout(layout, QHBoxLayout())
        self.gamry_location_button = Button("Select GamryCOM Location", self.choose_gamry_location)
        self.gamry_location_label.label().setWordWrap(True)
        self.default_location_button = Button(
            "Default Location",
            lambda: self.gamry_location_label.label().setText(gamry.DEFAULT_GAMRY_LOCATION),
        )
        layout_util.add_to_layout(
            button_layout, self.gamry_location_button, self.default_location_button
        )
        layout.addWidget(self.gamry_location_label)

        self.enabled_checkbox.stateChanged.connect(
            lambda checked: self.gamry_location_button.setEnabled(checked)
        )

        if self.enabled_checkbox.isChecked() and not GAMRY.is_valid():
            if OkCancelDialog(
                "Error",
                (
                    "Unable to load GamryCOM.\n"
                    f"{APP_NAME} will launch without Gamry features. You may change "
                    "this in the settings."
                ),
            ).run():
                self.enabled_checkbox.setChecked(False)
                self.save_on_close()
            else:
                sys.exit()

    def choose_gamry_location(self):
        """Select the location of GamryCOM."""
        filepath = QFileDialog.getOpenFileName(
            self,
            "Choose GamryCOM location",
            os.path.dirname(gamry.DEFAULT_GAMRY_LOCATION),
            "Executables (*.exe)",
        )[0]
        if filepath != "":
            self.gamry_location_label.label().setText(filepath)

    def save_on_close(self):
        """Call this when closing the settings window to save settings."""
        settings_dict = {
            gamry.ENABLED: self.enabled_checkbox.isChecked(),
            gamry.LOCATION: self.gamry_location_label.label().text(),
        }
        with open(gamry.SAVED_SETTINGS_FILE, "w") as f:
            json.dump(settings_dict, f)



from typing import Any, Self

from PyQt6.QtWidgets import QFormLayout

from .....classes import DescriptionInfo
from .....constants.paths.process import filenames
from .....custom_widgets import DoubleSpinBox
from .....instruments import INSTRUMENTS
from ...base_widget import AbstractBaseWidget
from . import encoding
from .process import IncrementTemperatureProcess


class IncrementTemperatureWidget(AbstractBaseWidget):
    """Increment the oven's temperature and wait for it to stabilize."""

    DISPLAY_NAME_PREFIX = "Increment Oven Temperature"

    def __init__(self):
        layout = QFormLayout()
        AbstractBaseWidget.__init__(
            self,
            layout,
            self.DISPLAY_NAME_PREFIX,
            IncrementTemperatureProcess,
            "thermometer--plus.png",
            DescriptionInfo(
                "temperature",
                "increment_temperature",
                IncrementTemperatureProcess.directory_name(),
                DescriptionInfo.Substitutions(
                    overview_dict={
                        "MEASUREMENT_INTERVAL": str(
                            IncrementTemperatureProcess.MEASUREMENT_INTERVAL
                        )
                    },
                    parameters_dict={"INCREMENT": encoding.INCREMENT},
                    data_recording_dict={
                        "TEMPERATURE_FILE": filenames.TEMPERATURES,
                        "GRAPH_FILE": IncrementTemperatureProcess.GRAPH_FILENAME,
                    },
                ),
            ),
        )

        oven = INSTRUMENTS.oven
        maximum_temperature = oven.maximum_temperature()
        self.increment_spinbox = DoubleSpinBox(
            oven.num_decimals(), -maximum_temperature, maximum_temperature
        )
        self.increment_spinbox.textChanged.connect(
            lambda value_as_str: self.setWindowTitle(
                f"{self.DISPLAY_NAME_PREFIX} ({value_as_str} degrees)"
            )
        )
        layout.addRow(encoding.INCREMENT, self.increment_spinbox)

    def to_dict(self) -> dict[str, Any]:
        return {encoding.INCREMENT: self.increment_spinbox.value()}

    @classmethod
    def from_dict(cls, data_as_dict: dict[str, Any]) -> Self:
        widget = cls()
        widget.increment_spinbox.setValue(data_as_dict[encoding.INCREMENT])
        return widget



class FatalSequenceError(BaseException):
    """A non-recoverable error for the sequence."""

    def __init__(self, error_message: str):
        BaseException.__init__(self)
        self.error_message = error_message


class PluginError(Exception):
    """An error caused by a faulty plugin."""



"""Filepaths and dictionary keys for oven settings."""

from .defaults import DEFAULT_SETTINGS
from .saved import SAVED_SETTINGS

SETTINGS_FOLDER_NAME = "oven"
SETTINGS_FILENAME = "oven_settings.json"

# files
SAVED_SETTINGS_FOLDER = SAVED_SETTINGS.joinpath(SETTINGS_FOLDER_NAME)
SAVED_SETTINGS_FILE = SAVED_SETTINGS.joinpath(SETTINGS_FILENAME)
DEFAULT_SETTINGS_FILE = DEFAULT_SETTINGS.joinpath(SETTINGS_FOLDER_NAME, SETTINGS_FILENAME)
OVEN_PORT_FILE = SAVED_SETTINGS.joinpath("oven_port.csv")
DESCRIPTION_FOLDER = DEFAULT_SETTINGS.joinpath(SETTINGS_FOLDER_NAME)
DESCRIPTION_FILENAME = "description.md"
# keys
TEMPERATURE_REGISTER = "Temperature Register"
SETPOINT_REGISTER = "Setpoint Register"
MAX_TEMPERATURE = "Maximum Temperature"
MIN_TEMPERATURE = "Minimum Temperature"
NUM_DECIMALS = "Number of Decimals"
STABILITY_TOLERANCE = "Stability Tolerance"
MEASUREMENT_INTERVAL = "Measurement Interval (ms)"
MINIMUM_STABILITY_MEASUREMENTS = "Minimum Measurements for Stability"
STABILITY_MEASUREMENT_INTERVAL = "Stability Check Interval (ms)"



from .button import BiggerButton, Button, FixedButton
from .combo_box import ComboBox
from .dialog import (
    Dialog,
    DontShowAgainDialog,
    OkCancelDialog,
    OkDialog,
    YesCancelDialog,
    YesNoDialog,
)
from .groupbox import GroupBox
from .label import IconLabel, Label
from .progressbar import ProgressBar
from .spin_box import DoubleSpinBox, SpinBox
from .tab_widget import TabWidget
from .widget import FixedWidget, SignalWidget, Widget



from __future__ import annotations

import asyncio
import contextlib
import copy
import json
import logging
import os
from asyncio import CancelledError
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, AsyncGenerator, Callable, Iterable

from PyQt6.QtCore import QObject, pyqtSignal

from ..constants.paths.sequence import METADATA_FILENAME
from .exceptions import FatalSequenceError
from .lock import DataLock
from .plotting import PlotHandle, PlotIndex, PlotSettings
from .sequence_step import SequenceStep

if TYPE_CHECKING:
    from ..tabs import SequenceDisplayTab


class StepRunner(QObject):
    """Runs `SequenceStep`s and provides utility functions that steps can call."""

    # title, message, options, response receiver
    promptRequested = pyqtSignal(str, str, dict, DataLock)
    """
    Emitted when a prompt needs to be sent the the main thread.

    Sends
    -----
    - The prompt title as a `str`.
    - The prompt text as a `str`.
    - The prompt options as a `dict[int, str]`.
    - The receiver as a `DataLock[int | None]`.
    """
    plotCommandRequested = pyqtSignal(object)
    """
    Emitted when a plot command needs to be run.

    Sends
    -----
    - The plot command as a `Callable[SequenceDisplayTab]`.
    """
    # the first argument is `qint64` because using `int` constrains it to 32 bits
    stepStateChanged = pyqtSignal("qint64", bool)
    """
    Emitted when a step is started/finished.

    Sends
    -----
    - The step's memory address as an `int`.
    - `True` if the step was started, `False` if it was finished (`bool`).
    """

    # ----------------------------------------------------------------------------------------------
    # public
    def __init__(self):
        QObject.__init__(self)

    async def run_steps(self, steps: Iterable[SequenceStep], data_directory: Path):
        """
        Run the provided steps sequentially. This can be called by `SequenceStep`s.

        Parameters
        ----------
        steps
            The steps to run.
        data_directory
            The base directory to put the steps' data directories in.

        Raises
        ------
        See `run_single_step()`.
        """
        for i, step in enumerate(steps):
            await self.run_single_step(step, data_directory, i + 1)  # run the step

    async def run_single_step(self, step: SequenceStep, data_directory: Path, step_number: int):
        """
        Run a single sequence step. This can be called by `SequenceStep`s.

        Parameters
        ----------
        step
            The `SequenceStep` to run.
        data_directory
            The base directory to put the step's data directory in.
        step_number
            The step's number (currently used to prefix the step's data directory).

        Raises
        ------
        CancelledError
            The sequence was cancelled.
        FatalSequenceError
            The sequence encountered a fatal error.

        Do not suppress these errors.
        """
        step_address = id(step)  # get the address
        # create the data directory first
        step_data_directory = await self.make_step_directory(data_directory, step, step_number)
        self.stepStateChanged.emit(step_address, True)  # notify start

        cancelled = False
        error_occurred = False
        start_datetime = datetime.now()  # record step start time
        try:
            try:
                await step.run(self, step_data_directory)
            except FatalSequenceError:  # fatal errors are not recoverable
                # intentionally putting this code here far clarity
                # we don't log metadata for fatal errors
                raise
            except Exception:  # recoverable error; log and ask the user what to do
                logging.getLogger(__name__).exception("Sequence step error")

                error_occurred = True
                # update the cancelled variable
                await self.prompt_handle_error(
                    step, f"The current step, {step.name()}, encountered an error."
                )
        except CancelledError:  # log cancellation then cancel
            cancelled = True
            await self.record_metadata(
                step_data_directory, step, start_datetime, cancelled, error_occurred
            )
            raise
        finally:
            self.stepStateChanged.emit(step_address, False)  # notify finish
        # this runs if there wasn't a fatal error and the sequence wasn't cancelled
        await self.record_metadata(
            step_data_directory, step, start_datetime, cancelled, error_occurred
        )

    async def prompt_user(self, step: SequenceStep, message: str, options: dict[int, str]) -> int:
        """
        Show a popup prompt to the user and wait for a response. This can be called by
        `SequenceStep`s.

        Parameters
        ----------
        step
            The `SequenceStep` requesting the prompt.
        message
            The prompt's text.
        options
            A mapping of values to options in the prompt. For example,
            `{1: "First Option", 2: "Second Option}` will show a prompt with options "First Option"
            and "Second Option". If the user selects "First Option", this function will return `1`.
            If the user selects "Second Option", this function will return `2`.
        """
        return await self.send_prompt_and_wait(
            f"Sequence: Message From {step.name()}", message, options
        )

    @contextlib.asynccontextmanager
    async def create_plot(
        self, step: SequenceStep, tab_text: str, plot_settings: PlotSettings
    ) -> AsyncGenerator[PlotHandle]:
        """
        Create a new plot on the visuals tab and return a handle to it. This waits until the plot
        is created. This can be called by `SequenceStep`s.

        Parameters
        ----------
        step
            The step creating the plot (generally just pass `self`).
        tab_text
            The text of the plot's tab.
        plot_settings
            The `PlotSettings` to configure the new plot with.

        Returns
        -------
        A thread-safe handle for the plot that can be used to modify it from your `SequenceStep`.
        """
        receiver: DataLock[PlotIndex | None] = DataLock(None)
        step_address = id(step)  # copy
        step_name = step.name()  # copy
        # notify that we want to create a new plot
        self.submit_plot_command(
            lambda plot_tab: plot_tab.add_plot(
                step_address, step_name, tab_text, plot_settings, receiver
            )
        )
        # wait for a response
        while True:
            await asyncio.sleep(0)
            if (plot_index := receiver.get()) is not None:
                plot_handle = PlotHandle(self, plot_index)
                try:
                    yield plot_handle  # return the plot handle
                finally:  # this runs at the end of the context manager
                    self.submit_plot_command(  # remove the plot
                        lambda plot_tab: plot_tab.remove_plot(copy.copy(plot_index))
                    )
                return  # exit

    # ----------------------------------------------------------------------------------------------
    # private
    async def make_step_directory(
        self, data_directory: Path, step: SequenceStep, number: int
    ) -> Path:
        """
        Create the data directory for **step** and return the directory's path.

        Raises
        ------
        CancelledError
            The sequence was cancelled.
        FatalSequenceError
            The sequence encountered a fatal error.
        """
        step_data_directory = data_directory.joinpath(f"{number} {step.directory_name()}")
        try:
            os.makedirs(step_data_directory, exist_ok=True)
        except OSError:
            if not await self.prompt_handle_error(
                step, f"Failed to create data directory {step_data_directory}"
            ):
                raise CancelledError
        return step_data_directory

    async def prompt_handle_error(self, step: SequenceStep, error_message: str):
        """
        Ask the user how to handle a sequence step encountering an error.

        Raises
        ------
        CancelledError
            The sequence was cancelled.
        FatalSequenceError
            An invalid response was sent (this indicates an issue with the core codebase).
        """
        response = await self.send_prompt_and_wait(
            f"Sequence: Error in {step.name()}",
            f"{error_message}\n\nSkip current step or cancel the sequence?",
            {0: "Skip Current", 1: "Cancel"},
        )
        match response:
            case 0:  # skip current
                return
            case 1:  # cancel
                raise CancelledError
            case _:  # this should never run
                raise FatalSequenceError(f"Reponse was {response} when it should have been 0 or 1")

    async def send_prompt_and_wait(self, title: str, message: str, options: dict[int, str]) -> int:
        """Private helper function to send a message the the user and wait for a response."""
        receiver: DataLock[int | None] = DataLock(None)
        self.promptRequested.emit(title, message, options, receiver)  # request a prompt
        # wait for the response
        while True:
            await asyncio.sleep(0)
            if (response := receiver.get()) is not None:
                return response

    async def record_metadata(
        self,
        directory: Path,
        step: SequenceStep,
        start_datetime: datetime,
        cancelled: bool,
        error_occurred: bool,
    ):
        """
        Generate the default metadata and combine it with the **step**'s metadata, then write the
        data to a metadata file in the **step**'s data directory. Logs any errors.

        Raises
        ------
        CancelledError
            The sequence was cancelled.
        FatalSequenceError
            The sequence encountered a fatal error.
        """
        while True:
            try:
                file = directory.joinpath(METADATA_FILENAME)
                data = step.metadata()
                # we do it in this order so the step's metadata gets overridden if there are
                # duplicate keys. The default keys are reserved
                data.update(
                    {
                        "Start Datetime": str(start_datetime),
                        "End Datetime": str(datetime.now()),
                        "Cancelled": cancelled,
                        "Error": error_occurred,
                    }
                )
                with open(file, "w") as f:
                    # any non-JSON types are converted to strings
                    json.dump(data, f, indent=4, default=str)
                break  # exit the loop
            except Exception:
                logging.getLogger(__name__).exception("Failed to write metadata")

                response = await self.prompt_user(
                    step,
                    "Failed to record metadata.\n\nRetry or cancel sequence?",
                    {0: "Retry", 1: "Cancel"},
                )
                match response:
                    case 0:  # retry
                        continue
                    case 1:  # cancel
                        raise CancelledError
                    case _:  # this should never run
                        raise FatalSequenceError(
                            f"Reponse was {response} when it should have been 0 or 1"
                        )

    # used externally
    def submit_plot_command(self, command: Callable[[SequenceDisplayTab], None]):
        """
        Submit a **command** to the plot command queue. This is not for direct use by
        `SequenceStep`s.
        """
        self.plotCommandRequested.emit(command)



from __future__ import annotations

from typing import TYPE_CHECKING

from PyQt6.QtWidgets import QMenu, QMenuBar

from ..classes import Action

if TYPE_CHECKING:
    from ..main_window import MainWindow


class ViewMenu(QMenu):
    """View menu."""

    def __init__(self, parent: QMenuBar, main_window: MainWindow):
        self.pop_graph: Action

        QMenu.__init__(self, "&View", parent)

        self.addAction(
            Action(main_window, "Fullscreen", main_window.toggle_fullscreen, shortcut="F11")
        )
        self.addAction(
            Action(
                main_window,
                "Shrink",
                main_window.shrink,
                shortcut="Ctrl+Shift+D",
            )
        )



from typing import Any

from .....classes import AbstractForegroundProcess, ProcessRunner
from .....instruments import Oven
from .....utility import dataframe
from ..set_temperature.encoding import SETPOINT
from ..set_temperature.process import SetTemperatureProcess
from . import encoding


class IncrementTemperatureProcess(SetTemperatureProcess):
    """Increment the oven's temperature and record data while waiting for it to stabilize."""

    def __init__(self, runner: ProcessRunner, data: dict[str, Any], name: str):
        data.update({SETPOINT: 0})
        SetTemperatureProcess.__init__(self, runner, data, name)

        self.increment = data[encoding.INCREMENT]

    @staticmethod
    def directory_name():
        return "Increment Temperature"

    def run(self):
        current_setpoint, proceed = self.get_setpoint()
        if not proceed:
            return
        setpoint, clamp_result = self.oven.clamp_setpoint(current_setpoint + self.increment)
        match clamp_result:
            case Oven.ClampResult.MAX_CLAMP | Oven.ClampResult.MIN_CLAMP:
                match clamp_result:
                    case Oven.ClampResult.MAX_CLAMP:
                        text = "Maximum"
                    case Oven.ClampResult.MIN_CLAMP:
                        text = "Minimum"
                self.communicate_error(f"{text} oven setpoint reached. The sequence will continue.")

        self.setpoint = setpoint
        SetTemperatureProcess.run(self)

    def title(self) -> str:  # overridden
        return f"Increment Temperature ({self.increment} °C)"

    def get_setpoint(self) -> tuple[float, bool]:
        """
        Get the oven's setpoint, going into an error state on failure. Retry until the read is
        successful.

        Returns
        -------
        A tuple of ([setpoint], [continue]). If [continue] is False, [setpoint] should be ignored.
        """
        setpoint = self.oven.get_setpoint()
        while setpoint is None:
            self.error_pause()
            if not self.wait(self.MEASUREMENT_INTERVAL, self.oven.is_connected):
                return (-1, False)
            setpoint = self.oven.get_setpoint()

        return (setpoint, True)

    def metadata(self):  # overridden
        return dataframe.add_to_dataframe(
            AbstractForegroundProcess.metadata(self), {"Selected Increment": self.increment}
        )



import logging
import tomllib
from abc import abstractmethod
from dataclasses import dataclass, field
from os import PathLike
from typing import Protocol

from jinja2 import Environment, FileSystemLoader, StrictUndefined, TemplateNotFound, UndefinedError

from ..constants.paths.descriptions import (
    DATA_RECORDING_FILENAME,
    OVERVIEW_FILENAME,
    PARAMETERS_FILENAME,
    VISUALS_FILENAME,
)
from ..constants.paths.process.filenames import METADATA_FILENAME

NO_DESCRIPTION_PROVIDED = "No description provided."
ERROR_TEXT = "Error loading description."

TEMPLATE = """{OVERVIEW}
# Parameters
{PARAMETERS}
# Visuals
{VISUALS}
# Data Recording
{DATA_RECORDING}"""

DATA_RECORDING_TEMPLATE = f"""Data is written to **{{DATA_DIRECTORY_NAME}}**, which contains:
- **{METADATA_FILENAME}**

    Metadata for this sequence step.
{{DATA_RECORDING_TEXT}}"""


class DescriptionProvider(Protocol):
    """An object that can provide a Markdown description string through its `render()` method."""

    @abstractmethod
    def render(self) -> str:
        """Generate the description."""
        ...


@dataclass
class TextDescription(DescriptionProvider):
    """
    Renders a description using text.

    Parameters
    ----------
    data_directory_name
        The name of the directory where the item records data.
    overview
        Text for the **Overview** section.
    parameters
        A mapping of `{"Parameter Name": "Description of parameter."}`. This will be rendered to a
        Markdown list for the **Parameters** section. Can be `{}` for no parameters.
    data_recording
        A mapping of `{"Filename": "Description of file contents."}`. This will be rendered to a
        Markdown list for the **Data Recording** section. Optional if your sequence step doesn't
        record extra data.
    visuals
        Text for the **Visuals** section. Optional if your sequence step doesn't have visuals.
    """

    data_directory_name: str
    overview: str
    parameters: dict[str, str]
    data_recording: dict[str, str] = field(default_factory=dict)
    visuals: str = "n/a"

    def render(self) -> str:  # implementation
        parameters_text = (
            generate_list_description(self.parameters) if len(self.parameters) > 0 else "n/a"
        )
        data_recording_text = (
            generate_list_description(self.data_recording) if len(self.data_recording) > 0 else ""
        )
        return TEMPLATE.format(
            OVERVIEW=self.overview,
            PARAMETERS=parameters_text,
            VISUALS=self.visuals,
            DATA_RECORDING=DATA_RECORDING_TEMPLATE.format(
                DATA_DIRECTORY_NAME=self.data_directory_name,
                DATA_RECORDING_TEXT=data_recording_text,
            ),
        )


@dataclass
class Substitutions:
    """
    Text-substitution dictionaries used by `jinja2`.

    Parameters
    ----------
    overview
        Substitutions for the **Overview** section.
    parameters
        Substitutions for the **Parameters** section.
    visuals
        Substitutions for the **Visuals** section.
    data_recording
        Substitutions for the **Data Recording** section.
    """

    overview: dict[str, str] = field(default_factory=dict)
    parameters: dict[str, str] = field(default_factory=dict)
    visuals: dict[str, str] = field(default_factory=dict)
    data_recording: dict[str, str] = field(default_factory=dict)


@dataclass
class FilesDescription(DescriptionProvider):
    """
    Renders a description by reading files in a folder.

    Parameters
    ----------
    description_folder
        The folder to read description files from. The folder should have the following structure:
    ```
    description_folder
    ├── data_recording.toml
    ├── overview.md
    ├── parameters.toml
    └── visuals.md
    ```
     For an example, see the [example directory](TODO). All files are mandatory, but only the
     overview file needs to have text (defaults will be used for the others).

    data_directory_name
        The name (i.e. not full path) of the directory that the item's process writes data to.
    substitutions
        Substitutions dictionaries used by `jinja2` when rendering the description files.
    """

    description_folder: PathLike[str] | str
    data_directory_name: str
    substitutions: Substitutions = field(default_factory=Substitutions)

    def render(self) -> str:  # implementation
        environment = Environment(
            loader=FileSystemLoader(self.description_folder), undefined=StrictUndefined
        )
        final_substitutions: dict[str, str] = {}

        # helper function to render a Markdown file
        def render_markdown(
            filename: str, substitution_dict: dict[str, str], empty_default: str
        ) -> str:
            try:
                text = environment.get_template(filename).render(substitution_dict)
                if text == "":
                    return empty_default
                return text
            except TemplateNotFound:
                return NO_DESCRIPTION_PROVIDED
            except UndefinedError:
                logging.getLogger(__name__).exception("Undefined jinja2 substitution")
                return ERROR_TEXT

        # helper function to render a TOML file into Markdown
        def render_toml(
            filename: str,
            substitution_dict: dict[str, str],
            empty_default: str,
            markdown_format_string: str = "**",
        ) -> str:
            try:
                name_description_map: dict[str, str] = tomllib.loads(
                    environment.get_template(filename).render(substitution_dict)
                )
                if len(name_description_map) == 0:
                    return empty_default
                return generate_list_description(name_description_map, markdown_format_string)
            except Exception:
                logging.getLogger(__name__).exception("Failed to render TOML description file")
                raise

        # overview
        final_substitutions["OVERVIEW"] = render_markdown(
            OVERVIEW_FILENAME, self.substitutions.overview, NO_DESCRIPTION_PROVIDED
        )
        # visuals
        final_substitutions["VISUALS"] = render_markdown(
            VISUALS_FILENAME, self.substitutions.visuals, "n/a"
        )
        # parameters
        try:
            parameters_text = render_toml(PARAMETERS_FILENAME, self.substitutions.parameters, "n/a")
        except TemplateNotFound:
            parameters_text = NO_DESCRIPTION_PROVIDED
        except Exception:
            parameters_text = ERROR_TEXT
        final_substitutions["PARAMETERS"] = parameters_text
        # data recording
        try:
            data_recording_text = DATA_RECORDING_TEMPLATE.format(
                DATA_DIRECTORY_NAME=self.data_directory_name,
                DATA_RECORDING_TEXT=render_toml(
                    DATA_RECORDING_FILENAME, self.substitutions.data_recording, ""
                ),
            )
        except TemplateNotFound:
            data_recording_text = NO_DESCRIPTION_PROVIDED
        except Exception:
            data_recording_text = ERROR_TEXT
        final_substitutions["DATA_RECORDING"] = data_recording_text

        return TEMPLATE.format(**final_substitutions)


def generate_list_description(
    name_description_map: dict[str, str], markdown_format_string: str = "**"
) -> str:
    """
    Creates a Markdown list string.

    Parameters
    ----------
    name_description_map
        A mapping of parameter names to their description.
    markdown_format_string
        The format string to place on both sides of the parameter name. For example, `**` will make
        the parameter name bold.

    Examples
    --------
    >>> generate_list_description(
    ...     {
    ...         "Parameter1": "Description of Parameter1.",
    ...         "Parameter2": "Description of Parameter2."
    ...     },
    ...     "**",
    ... )
    When rendered with Markdown, this yields:

    - **Parameter1**

        Description of Parameter1.
    - **Parameter2**

        Description of Parameter2.
    """
    return "\n".join(
        f"- {markdown_format_string}{name}{markdown_format_string}\n\n    {description}"
        for name, description in name_description_map.items()
    )



import json
import os
import typing
from pathlib import Path
from types import ModuleType
from typing import Mapping

from PyQt6.QtCore import QSize
from PyQt6.QtWidgets import QFileDialog, QGridLayout, QHBoxLayout, QSizePolicy, QStackedWidget

from ..classes import SequenceRunner
from ..constants.paths.settings import sequence as sequence_paths
from ..custom_widgets import Button, IconLabel, Label, Widget
from ..enums import SequenceStatus
from ..menu import SequenceMenu
from ..sequence_builder import OptionsTreeWidget, SequenceTreeWidget
from ..utility import errors, images
from .sequence_display import SequenceDisplayTab


class SequenceBuilderTab(Widget):
    """
    Sequence tab.

    Parameters
    ----------
    visuals_tab
        The tab where plots are shown for sequence steps.
    menu
        The `QMenu` that holds sequence controls.
    plugin_modules
        A list of loaded plugin modules.
    """

    def __init__(
        self,
        visuals_tab: SequenceDisplayTab,
        menu: SequenceMenu,
        plugin_modules: Mapping[str, ModuleType],
    ):
        layout = QGridLayout()
        Widget.__init__(self, layout)

        self.sequence_runner: SequenceRunner | None = None  # have to keep a reference to the runner

        self.visuals_tab = visuals_tab  # another tab
        self.menu = menu

        self.options_tree = OptionsTreeWidget.from_plugins(plugin_modules)
        self.sequence_tree = SequenceTreeWidget.from_autosave()

        self.directory_button = Button("Choose Data Directory", self.choose_directory)
        self.directory_label = IconLabel(
            images.make_pixmap("folder--arrow.png"), self.load_previous_directory()
        )

        self.button_layout = QStackedWidget()
        self.start_button = Button("Start", self.start_sequence)
        self.pause_button = Button("Pause")
        self.unpause_button = Button("Unpause")
        self.status_label = Label("Inactive").set_color("gray")

        self.arrange_widgets(layout)

    def arrange_widgets(self, layout: QGridLayout):
        """Arrange widgets at construction."""
        layout.addWidget(self.options_tree, 0, 0)
        layout.addWidget(self.sequence_tree, 0, 1)
        layout.setRowStretch(0, 1)

        BUTTON_SIZE = QSize(250, 50)

        # data directory widgets
        directory_layout = QHBoxLayout()
        layout.addLayout(directory_layout, 1, 0)
        self.directory_button.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Preferred)
        self.directory_button.setMinimumSize(BUTTON_SIZE)
        self.directory_label.label().setWordWrap(True)
        directory_layout.addWidget(self.directory_button)
        directory_layout.addWidget(self.directory_label)

        # start/pause button and status label
        runner_layout = QHBoxLayout()
        layout.addLayout(runner_layout, 1, 1)
        # buttons
        self.button_layout.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Preferred)
        self.button_layout.setMinimumSize(BUTTON_SIZE)
        runner_layout.addWidget(self.button_layout)
        self.start_button.setEnabled(self.directory_is_valid())
        for button in [self.start_button, self.pause_button, self.unpause_button]:
            self.button_layout.addWidget(button)
        # label
        font = self.status_label.font()
        font.setPointSize(16)
        self.status_label.setFont(font)
        runner_layout.addWidget(self.status_label)

    def choose_directory(self):
        """Open a dialog to choose the data-storage directory."""
        directory = QFileDialog.getExistingDirectory(
            self,
            "Select sequence data location",
            os.path.expanduser("~"),
            QFileDialog.Option.ShowDirsOnly,
        )
        self.directory_label.label().setText(directory)  # update label

        self.start_button.setEnabled(directory != "")  # enable/disable start button

    def directory_is_valid(self):
        """Whether the selected directory is valid."""
        return self.directory_label.label().text() != ""

    def load_previous_directory(self) -> str:
        """Try to load the previously used directory."""
        try:
            with open(sequence_paths.SEQUENCE_DIRECTORY_FILE, "r") as f:
                directory = typing.cast(str, json.load(f))
                return directory
        except Exception:
            return ""

    def data_directory(self) -> str:
        """Get the current data directory."""
        return self.directory_label.label().text()

    def save_on_close(self):
        """Call this when closing the application to save settings."""
        self.options_tree.save_on_close()
        self.sequence_tree.save_on_close()

        directory = self.data_directory()
        try:
            with open(sequence_paths.SEQUENCE_DIRECTORY_FILE, "w") as f:
                json.dump(directory, f)
        except OSError:
            pass

    # ----------------------------------------------------------------------------------------------
    # sequence
    def start_sequence(self):
        self.sequence_runner = SequenceRunner()  # store a reference
        self.pause_button.clicked.connect(self.sequence_runner.pause)
        self.unpause_button.clicked.connect(self.sequence_runner.unpause)
        self.menu.cancel.triggered.connect(self.sequence_runner.cancel)
        if self.sequence_runner.run_sequence(
            self, self.sequence_tree.view.model(), Path(self.data_directory())
        ):
            self.handle_sequence_state_change(True)

    def handle_sequence_status_change(self, status: SequenceStatus):
        """Handle the sequence's status changing."""
        match status:
            case SequenceStatus.Active:
                text = "Active"
                color = "green"
                button = self.pause_button

            case SequenceStatus.Paused:
                text = "Paused"
                color = "cyan"
                button = self.unpause_button

            case SequenceStatus.Completed:
                text = "Completed"
                color = "gray"
                button = self.start_button

            case SequenceStatus.Cancelled:
                text = "Cancelled"
                color = "gray"
                button = self.start_button

            case SequenceStatus.FatalError:
                text = "Fatal Error"
                color = "red"
                button = self.start_button
                # notify the user
                errors.show_error_delayed(
                    "Sequence: Fatal Error",
                    "The sequence encountered a fatal error and was terminated. "
                    "See the error log for details.",
                )
            case _:  # this should never run
                raise ValueError(f"Unknown `SequenceStatus`: {status}")

        self.button_layout.setCurrentWidget(button)
        self.status_label.setText(text)
        self.status_label.set_color(color)

    def handle_sequence_state_change(self, running: bool):
        """Handle the sequence starting/stopping."""
        self.menu.cancel.setEnabled(running)
        self.directory_button.setEnabled(not running)
        self.start_button.setEnabled(not running)  # make sure the user can't spam the start button
        if not running:
            self.sequence_runner = None  # delete the runner

        # TODO: enable/disable the model



from .....gamry_integration import GAMRY

INITIAL_FREQUENCY = "Initial Frequency (Hz)"
FINAL_FREQUENCY = "Final Frequency (Hz)"
POINTS_PER_DECADE = "Points Per Decade"
AC_VOLTAGE = "AC Voltage (mV rms)"
DC_VOLTAGE = "DC Voltage (V)"
DC_VOLTAGE_REFERENCE = "DC Voltage Reference"
ESTIMATED_IMPEDANCE = "Estimated Z (ohms)"
IMPEDANCE_READER_SPEED = "Impedance Reader Speed"
SELECTED_PSTATS = "Potentiostat(s)"

VS_EREF = "Vs. Eref"
VS_EOC = "Vs. Eoc"

# fails if GamryCOM is not enabled or could not be loaded
try:
    IMPEDANCE_READER_SPEED_DICT: dict[str, int] = {
        "Fast": GAMRY.com_interface().ReadZSpeedFast,
        "Normal": GAMRY.com_interface().ReadZSpeedNorm,
        "Low Noise": GAMRY.com_interface().ReadZSpeedLow,
    }
except Exception:
    IMPEDANCE_READER_SPEED_DICT = dict()


class Headers:
    """Headers for EIS files."""

    # these are tab-separated
    GAMRY_HEADERS = (
        ("EXPLAIN",),
        ("TAG", "EISPOT"),
        ("TITLE", "LABEL", "Potentiostatic EIS", "Test &Identifier"),
        (),
    )
    DATA_HEADERS = (
        ("ZCURVE", "TABLE"),
        (
            "",  # empty string to start line with tab
            "Pt",
            "Time",
            "Freq",
            "Zreal",
            "Zimag",
            "Zsig",
            "Zmod",
            "Zphz",
            "Idc",
            "Vdc",
            "IERange",
        ),
        ("", "#", "s", "Hz", "ohm", "ohm", "V", "ohm", "°", "A", "V", "#"),
    )


class FileFormat:
    DELIMETER = "\t"



from PyQt6.QtWidgets import QFormLayout, QHBoxLayout, QVBoxLayout

from ..constants.paths.settings import oven as settings
from ..custom_widgets import ComboBox, GroupBox, Label
from ..enums import ConnectionStatus
from ..utility import layout as layout_util
from . import ports


class InstrumentConnectionWidget(GroupBox):
    """Widget for changing the instrument connection ports."""

    MAX_COMBOBOX_ITEMS = 500
    NULL_TEXT = "------------"

    def __init__(self):
        GroupBox.__init__(self, "Instrument Connection", QHBoxLayout())
        self.oven = INSTRUMENTS.oven

        self.create_widgets()

        # check the oven's connection on an interval
        self.connect_signals()
        self.oven.start()
        self.load_ports()

    def create_widgets(self):
        """Create subwidgets."""
        layout: QHBoxLayout = self.layout()  # type: ignore

        self.oven_combobox = ComboBox()
        # add all available COM ports to the list
        self.oven_combobox.addItems(ports.list_ports())
        self.oven_connection_label = Label(self.NULL_TEXT)

        # this could probably be put in a for loop if more instruments are added
        inner_layout = layout_util.add_sublayout(layout, QVBoxLayout())
        # the top label and combobox
        layout_util.add_to_layout(inner_layout, Label("Oven Port"), self.oven_combobox)
        # the two bottom labels with the connection status
        label_layout = layout_util.add_sublayout(inner_layout, QFormLayout())
        label_layout.addRow("Status:", self.oven_connection_label)

    def connect_signals(self):
        """Give widgets logic."""
        # changing the oven combobox updates the oven port instantly
        self.oven_combobox.activated.connect(self.update_port)
        self.oven_combobox.pressed.connect(self.update_comboboxes)
        self.oven.connectionChanged.connect(self.handle_connection_change)

    def update_comboboxes(self):
        """Update the port comboboxes to show the ports that currently exist."""
        text = self.oven_combobox.currentText()
        self.oven_combobox.blockSignals(True)
        self.oven_combobox.clear()
        self.oven_combobox.addItems(ports.list_ports())
        self.oven_combobox.setCurrentText(text)
        self.oven_combobox.blockSignals(False)
        # NOTE: when adding additional instruments, make sure they can never use the same port

    def update_connection_label(self, connected: bool):
        """Update the connection status label's text and color."""
        text = ConnectionStatus.bool_to_str(connected)
        self.oven_connection_label.setText(text)
        # this is HTML syntax
        self.oven_connection_label.setStyleSheet(
            "color: " + ConnectionStatus.bool_to_color(connected)
        )

    def handle_connection_change(self, connected: bool):
        self.update_connection_label(connected)

    def update_port(self, index: int | None = None):
        """Update the oven's port (this is a slot)."""
        self.oven.set_port(self.oven_combobox.currentText())

    def save_on_close(self):
        """Call this when closing the application to save the oven port."""
        port = self.oven_combobox.currentText()
        with open(settings.OVEN_PORT_FILE, "w") as f:
            f.write(port)

    def load_ports(self):
        """Loads saved port selections."""
        try:
            with open(settings.OVEN_PORT_FILE, "r") as f:
                port = f.read().strip()
            # if the combobox contains the previously stored port
            if self.oven_combobox.findText(port) != -1:
                self.oven_combobox.setCurrentText(port)
                self.oven.set_port(port)
        except Exception:
            pass



from __future__ import annotations

import importlib
import logging
import pkgutil
from importlib import metadata
from types import ModuleType

from .. import plugins
from ..constants import PLUGIN_ENTRY_POINT


def load_plugins_from_module(module: ModuleType) -> tuple[dict[str, ModuleType], list[str]]:
    """
    Load plugins that are submodules of **module**.

    Returns
    -------
    A tuple of (the successfully loaded plugin modules, the modules that failed).
    """
    failure_plugins: list[str] = []
    plugin_modules: dict[str, ModuleType] = {}
    # search for packages in the `plugins` directory
    for _, name, is_package in pkgutil.iter_modules(module.__path__):
        if not is_package:  # ignore non packages
            continue
        try:
            plugin_module = importlib.import_module(module.__name__ + "." + name)
            plugin_modules[name] = plugin_module
        except Exception:
            logging.getLogger(__name__).exception(f"Failed to load local plugin {name}")
            failure_plugins.append(name)

    return (plugin_modules, failure_plugins)


def load_local_plugins() -> tuple[dict[str, ModuleType], list[str]]:
    """Load local plugins from the `plugins` directory."""
    # see `load_plugins_from_module()` for return value
    return load_plugins_from_module(plugins)


def load_global_plugins() -> tuple[dict[str, ModuleType], list[str]]:
    """
    Load plugins installed in the current Python environment (i.e. plugins installed with `pip`).

    Returns
    -------
    A tuple of (the successfully loaded plugin modules, the modules that failed).
    """
    failure_plugins: list[str] = []
    plugin_modules: dict[str, ModuleType] = {}
    for entry_point in metadata.entry_points(group=PLUGIN_ENTRY_POINT):
        name = entry_point.module
        try:
            plugin_modules[name] = entry_point.load()
        except Exception:
            logging.getLogger(__name__).exception(f"Failed to load global plugin {name}")
            failure_plugins.append(name)

    return (plugin_modules, failure_plugins)


def load_all_plugins() -> tuple[dict[str, ModuleType], list[str], list[str]]:
    """
    Load plugins for the application.

    Returns
    -------
    A tuple of (the successfully loaded plugin modules, the global plugins that failed, the local
    plugins that failed).

    Notes
    -----
    If any duplicate plugin names are found, the environment-installed (global) plugin is used.
    """
    local_plugin_modules, local_failure_plugins = load_local_plugins()
    global_plugin_modules, global_failure_plugins = load_global_plugins()

    plugin_modules = local_plugin_modules
    # fuse the plugin lists. If there are any duplicate plugin names, environment-installed plugin
    # is used
    plugin_modules.update(global_plugin_modules)

    return (plugin_modules, global_failure_plugins, local_failure_plugins)



"""File paths for the application."""

from . import descriptions, icons, settings
from .base import ASSETS, PLUGINS

FOLDERS_TO_CREATE = [
    settings.saved.SAVED_SETTINGS,
    settings.oven.SAVED_SETTINGS_FOLDER,
    settings.sequence.SAVED_SETTINGS_FOLDER,
    settings.gamry.SAVED_SETTINGS_FOLDER,
]



from .augmented import (
    BiggerButton,
    Button,
    ComboBox,
    Dialog,
    DontShowAgainDialog,
    DoubleSpinBox,
    FixedButton,
    FixedWidget,
    GroupBox,
    IconLabel,
    Label,
    OkCancelDialog,
    OkDialog,
    ProgressBar,
    SignalWidget,
    SpinBox,
    TabWidget,
    Widget,
    YesCancelDialog,
    YesNoDialog,
)
from .container import Container, FixedContainer
from .markdown_view import MarkdownView
from .parameter_description import ParameterDescriptionWidget
from .plot import PlotWidget
from .separator import HSeparator, VSeparator



from typing import Self

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QPixmap
from PyQt6.QtWidgets import QHBoxLayout, QLabel, QSizePolicy

from ..container import Container


class Label(QLabel):
    """QLabel with the ability to select text."""

    def __init__(self, text: str = ""):
        QLabel.__init__(self, text)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)

    def set_color(self, color: str) -> Self:
        """Set the label's text color. Supports color names, like "blue" or "lightgreen"."""
        self.setStyleSheet("color: " + color)
        return self

    def reset(self, reset_text: str) -> Self:
        """Reset the label's color and set the text to **reset_text**."""
        self.set_color("")
        self.setText(reset_text)
        return self


class IconLabel(Container):
    """Label with an icon on the left. Both are contained in a `Container` (this widget)."""

    def __init__(self, icon: QPixmap, text: str = ""):
        layout = QHBoxLayout()
        Container.__init__(self, layout)

        self.icon_label = QLabel()
        self.icon_label.setPixmap(icon)
        self.text_label = Label(text)

        layout.addWidget(self.icon_label)
        layout.addWidget(self.text_label)

        self.icon_label.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)

    def label(self) -> Label:
        """Get the text label."""
        return self.text_label

    def pixmap(self) -> QPixmap:
        """Get the pixmap."""
        return self.icon_label.pixmap()



"""The base directory."""

import sys
from importlib import resources
from pathlib import Path

from ..name import PACKAGE_NAME

with resources.path(PACKAGE_NAME) as path:
    if not path.exists():  # this happens if we're frozen (packaged)
        # the path to the folder containing the executable
        FOLDER = Path(sys._MEIPASS)  # type: ignore
    else:
        FOLDER = path

ASSETS = FOLDER.joinpath("assets")
PLUGINS = FOLDER.joinpath("plugins")



from abc import abstractmethod
from typing import Iterable, Protocol

from PyQt6.QtGui import QIcon

from ..classes import SequenceStep
from ..utility.serde import Deserialize, Serialize
from .item_widget import ItemWidget


class DataItem(Deserialize, Serialize, Protocol):
    """An inner item used by a `SequenceItem`."""

    # visuals
    @abstractmethod
    def display_name(self) -> str:
        """Get the name displayed on the item."""
        ...

    @abstractmethod
    def icon(self) -> QIcon:
        """Get the icon to display on the item."""
        ...

    @abstractmethod
    def open_event(self, editable: bool):
        """
        Handle the item being "opened" (i.e. double clicked).

        Parameters
        ----------
        editable
            Whether the item's data should be editable.
        """
        ...

    def supports_subitems(self) -> bool:
        """Whether the item supports subitems. By default, items do not support subitems."""
        return False

    def expand_event(self):
        """Handle the item being expanded. Does nothing by default."""
        return

    def collapse_event(self):
        """Handle the item being collapsed. Does nothing by default."""
        return

    # sequence
    @abstractmethod
    def create_sequence_step(self, substeps: Iterable[SequenceStep]) -> SequenceStep:
        """
        Create the `SequenceStep` this item represents.

        Parameters
        ----------
        substeps
            The `SequenceStep`s created from this item's subitems. For items that do not support
            subitems, this is always an empty iterable.

        Returns
        -------
        This item's `SequenceStep`.
        """
        ...


class WidgetDataItem(DataItem, Protocol):
    """A `DataItem` that uses an `ItemWidget` to display and modify data."""

    @abstractmethod
    def widget(self) -> ItemWidget:  # private
        """Get the widget this item uses."""
        ...

    def display_name(self) -> str:  # implementation
        return self.widget().windowTitle()

    def icon(self) -> QIcon:  # implementation
        return self.widget().windowIcon()

    def open_event(self, editable: bool):  # implementation
        self.widget().show_editable(editable)



import json
from os import PathLike

from PyQt6.QtWidgets import QCheckBox, QMessageBox


class Dialog(QMessageBox):
    """
    Base class for dialog pop-ups.

    Parameters
    ----------
    title
        The window title.
    message
        The message text.
    buttons
        The buttons used on the dialog.
    default_button
        The initially selected button. Defaults to `StandardButton.Ok` if the dialog has that
        button, otherwise no button is selected.
    """

    def __init__(
        self,
        title: str,
        message: str,
        buttons: QMessageBox.StandardButton,
        default_button: QMessageBox.StandardButton = QMessageBox.StandardButton.Ok,
    ):
        QMessageBox.__init__(self)

        self.setWindowTitle(title)
        self.setText(message)
        self.setStandardButtons(buttons)
        self.setDefaultButton(default_button)

    def run(self) -> bool:
        """Show the dialog and return whether the proposed action was accepted."""
        self.exec()
        role = self.buttonRole(self.clickedButton())
        Roles = QMessageBox.ButtonRole
        match role:
            case Roles.AcceptRole | Roles.YesRole | Roles.ApplyRole:
                return True
        return False


class YesCancelDialog(Dialog):
    """Dialog pop-up with Yes and Cancel buttons."""

    def __init__(self, title: str, message: str):
        Dialog.__init__(
            self,
            title,
            message,
            Dialog.StandardButton.Yes | Dialog.StandardButton.Cancel,
            Dialog.StandardButton.Yes,
        )


class OkDialog(Dialog):
    """Dialog pop-up with an Ok button. `run()` always returns `True`."""

    def __init__(self, title: str, message: str):
        Dialog.__init__(self, title, message, Dialog.StandardButton.Ok, Dialog.StandardButton.Ok)


class OkCancelDialog(Dialog):
    """Dialog pop-up with Ok and Cancel buttons."""

    def __init__(self, title: str, message: str):
        Dialog.__init__(
            self,
            title,
            message,
            Dialog.StandardButton.Ok | Dialog.StandardButton.Cancel,
            Dialog.StandardButton.Ok,
        )


class YesNoDialog(Dialog):
    """Dialog pop-up with Yes and No buttons."""

    def __init__(self, title: str, message: str):
        Dialog.__init__(
            self,
            title,
            message,
            Dialog.StandardButton.Yes | Dialog.StandardButton.No,
            Dialog.StandardButton.Yes,
        )


class DontShowAgainDialog(Dialog):
    """
    A dialog that has a "Don't show again" option.

    Parameters
    ----------
    title
        The title of the dialog.
    message
        The message text shown on the dialog.
    buttons
        `StandardButton`s to show on the dialog.
    default_button
        Which `StandardButton` to have selected by default.
    file
        The JSON file where the "don't show again" state is saved to and loaded from.
    """

    def __init__(
        self,
        title: str,
        message: str,
        buttons: QMessageBox.StandardButton,
        default_button: QMessageBox.StandardButton,
        file: PathLike[str] | str,
    ):

        Dialog.__init__(self, title, message, buttons, default_button)
        self.check_box = QCheckBox("Don't show again", self)
        self.setCheckBox(self.check_box)

        self.file = file
        try:
            with open(self.file, "r") as f:
                if not json.load(f):
                    self.check_box.setChecked(True)
        except OSError:
            pass

    def should_run(self) -> bool:
        """
        Whether we should even run the dialog (we shouldn't run if the user checked "Don't show
        again").
        """
        return not self.check_box.isChecked()

    def save_state(self):
        """Save the state of the "Don't show again" checkbox to a file."""
        try:
            with open(self.file, "w") as f:
                json.dump(not self.check_box.isChecked(), f)
        except OSError:
            pass

    def run(self) -> bool:
        """
        Show the dialog and save the state of the "Don't show again" checkbox afterward. Returns
        whether the proposed action was accepted. If the "Don't show again" checkbox is checked,
        this always returns `True` without showing the dialog.
        """
        if self.should_run():
            result = Dialog.run(self)
            self.save_state()
            return result
        return True



import json
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Union

import minimalmodbus as modbus
from PyQt6.QtCore import QMutex, QMutexLocker, QObject, QThread, pyqtSignal

from .classes import ABCQObjectMeta, DataMutex
from .classes.timer import Timer
from .constants.paths import settings
from .enums import ConnectionStatus

if TYPE_CHECKING:
    from developer import DeveloperOven


class Instrument(QObject, metaclass=ABCQObjectMeta):
    """
    Abstract class to represent instruments. When using methods that change the physical oven,
    call `acquire()` followed at some point by `release()`.
    """

    # signals
    connectionChanged = pyqtSignal(bool)
    """
    Fires whenever the connection status changes. Sends whether the instrument is connected as a
    `bool`.
    """
    lockChanged = pyqtSignal(bool)
    """
    Fires whenever the instrument is locked/unlocked. Sends whether the instrument is unlocked as a
    `bool`.
    """

    def __init__(self):
        QObject.__init__(self)
        self.connection_status = ConnectionStatus.NULL
        # this lock system indicates to the rest of the application whether the instrument is
        # available
        self.unlocked = DataMutex(True)
        # this lock ensures accessing the physical device is thread safe
        self.physical_device_lock = QMutex()

    def lock(self):
        """
        Make the instrument available for only one process. This should be followed by a call to
        `unlock()`. Emits signals.
        """
        self.unlocked.set(False)
        self.lockChanged.emit(False)

    def unlock(self):
        """Make the instrument available for other processes. Emits signals."""
        self.unlocked.set(True)
        self.lockChanged.emit(True)

    def is_unlocked(self) -> bool:
        """Whether the instrument is available."""
        return self.unlocked.get()

    def device_lock(self) -> QMutex:
        """Get the device lock, which ensures accessing the physical device is thread safe."""
        return self.physical_device_lock

    def update_connection_status(self, connection_status: ConnectionStatus):
        """
        Helper function to update the connection status and emit a signal. This should not be
        called outside of the oven.
        """
        if self.connection_status != connection_status:
            self.connection_status = connection_status
            self.connectionChanged.emit(self.is_connected())

    @abstractmethod
    def connect(self):
        """Connect to the physical device."""
        pass

    def is_connected(self) -> bool:
        """Get the connection status of this instrument as a `bool`."""
        return bool(self.connection_status)


class InstrumentLocker[InstrumentType: Instrument]:  # generic
    """Locks and unlocks an instrument using the context manager."""

    def __init__(self, instrument: InstrumentType):
        self.instrument = instrument

    def __enter__(self) -> InstrumentType:
        self.instrument.lock()
        return self.instrument

    def __exit__(self, *exception_info):
        # we don't care about exceptions here
        self.instrument.unlock()


class Oven(Instrument):
    """
    Class to represent the physical oven the application controls. Once `start()` is called, the
    oven will attempt to connect automatically and will also run a stability check.
    """

    # signals
    temperatureChanged = pyqtSignal(float)
    """Fires whenever the oven's temperature *changes*. Sends the temperature as a `float`."""
    setpointChanged = pyqtSignal(float)
    """Fires whenever the oven's setpoint *changes*. Sends the setpoint as a `float`."""
    stabilityChanged = pyqtSignal(bool)
    """
    Fires whenever the oven's stability status changes. Sends a `bool` indicating whether the
    temperature is stable.
    """
    stabilityCountChanged = pyqtSignal(int)
    """
    Fires whenever the oven's stability measurement count changes. Sends the new count as an `int`.
    """

    class ClampResult(Enum):
        """Result enum used by the `clamp_setpoint()` method."""

        NO_CLAMP = 0
        """No clamping occurred."""
        MAX_CLAMP = 1
        """The setpoint was clamped to the maximum oven temperature."""
        MIN_CLAMP = 2
        """The setpoint was clamped to the minimum oven temperature."""

    def __init__(self, oven_port: str = ""):
        Instrument.__init__(self)
        self.port = oven_port
        # initialize the oven settings
        self.max_temperature: float
        self.min_temperature: float
        self.setpoint_register: int
        self.temperature_register: int
        self.number_of_decimals: int
        self.stability_tolerance: float
        self.minimum_stability_measurements: int
        self.measurement_interval: int
        self.stability_measurement_interval: int
        self.load_settings()
        # do not access these directly
        self.last_temperature: float | None = None
        self.last_setpoint: float | None = None
        self.stability_measurement_count = 0
        self.stable = DataMutex(False)
        self.disconnected_count = 0
        # thread for continuous background tasks like monitoring the temperature
        self.task_thread = QThread()

    def load_settings(self):
        """Load the oven settings from the oven settings file."""
        try:
            with open(settings.oven.SAVED_SETTINGS_FILE, "r") as f:
                saved_settings = json.load(f)
        except Exception:
            with open(settings.oven.DEFAULT_SETTINGS_FILE, "r") as f:
                saved_settings = json.load(f)
        SettingNames = settings.oven
        self.max_temperature = saved_settings[SettingNames.MAX_TEMPERATURE]
        self.min_temperature = saved_settings[SettingNames.MIN_TEMPERATURE]
        self.setpoint_register = saved_settings[SettingNames.SETPOINT_REGISTER]
        self.temperature_register = saved_settings[SettingNames.TEMPERATURE_REGISTER]
        self.number_of_decimals = saved_settings[SettingNames.NUM_DECIMALS]
        self.stability_tolerance = saved_settings[SettingNames.STABILITY_TOLERANCE]
        self.minimum_stability_measurements = saved_settings[
            SettingNames.MINIMUM_STABILITY_MEASUREMENTS
        ]
        self.measurement_interval = saved_settings[SettingNames.MEASUREMENT_INTERVAL]
        self.stability_measurement_interval = saved_settings[
            SettingNames.STABILITY_MEASUREMENT_INTERVAL
        ]

    def maximum_temperature(self) -> float:
        """Get the oven's maximum allowed temperature."""
        return self.max_temperature

    def minimum_temperature(self) -> float:
        """Get the oven's minimum allowed temperature."""
        return self.min_temperature

    def minimum_measurements(self) -> int:
        """Get the oven's minimum number of measurements for stability."""
        return self.minimum_stability_measurements

    def num_decimals(self) -> int:
        """Get the oven's number of decimals."""
        return self.number_of_decimals

    def read_temp(self) -> float | None:
        """
        Returns the oven's temperature if on a successful read, `None` otherwise. This is
        thread-safe.
        """
        with QMutexLocker(self.device_lock()):
            try:
                temperature = float(
                    self.device.read_register(self.temperature_register, self.number_of_decimals)
                )
                self.update_temperature(temperature)
                return temperature
            except Exception:
                self.update_connection_status(ConnectionStatus.DISCONNECTED)
                return None

    def change_setpoint(self, setpoint: float) -> bool:
        """
        Sets the oven's temperature to **setpoint**. Returns `True` if the operation was
        successful, `False` otherwise. If the provided setpoint is outside the oven's setpoint
        range, it is clamped. This reset's the oven's stability. This is thread-safe.
        """
        with QMutexLocker(self.device_lock()):
            try:
                self.device.write_register(
                    self.setpoint_register,
                    self.clamp_setpoint(setpoint)[0],
                    self.number_of_decimals,
                )
                self.update_setpoint(setpoint)
                self.reset_stability()
                return True
            except Exception:
                self.update_connection_status(ConnectionStatus.DISCONNECTED)
                return False

    def get_setpoint(self) -> float | None:
        """Get the oven's setpoint. Returns `None` if the read fails. This is thread-safe."""
        with QMutexLocker(self.device_lock()):
            try:
                setpoint = float(
                    self.device.read_register(self.setpoint_register, self.number_of_decimals)
                )
                self.update_setpoint(setpoint)
                return setpoint
            except Exception:
                self.update_connection_status(ConnectionStatus.DISCONNECTED)
                return None

    def update_temperature(self, temperature: float):
        """Update the last temperature and emit signals on a change. This method is private."""
        if self.last_temperature is None or self.last_temperature != temperature:
            self.last_temperature = temperature
            self.temperatureChanged.emit(temperature)

    def update_setpoint(self, setpoint: float):
        """Update the last setpoint and emit signals on a change. This method is private."""
        if self.last_setpoint is None or self.last_setpoint != setpoint:
            self.last_setpoint = setpoint
            self.setpointChanged.emit(setpoint)

    def clamp_setpoint(self, setpoint: float) -> tuple[float, ClampResult]:
        """Clamp the provided **setpoint** to be within the oven's physical limits."""
        if setpoint > self.max_temperature:
            return (self.max_temperature, Oven.ClampResult.MAX_CLAMP)
        elif setpoint < self.min_temperature:
            return (self.min_temperature, Oven.ClampResult.MIN_CLAMP)
        return (setpoint, Oven.ClampResult.NO_CLAMP)

    def connect(self):
        """Attempts to connect to the oven. This is thread-safe."""
        with QMutexLocker(self.device_lock()):
            try:
                self.device = modbus.Instrument(self.port, 1, close_port_after_each_call=True)
                connection_status = ConnectionStatus.CONNECTED
            except Exception:
                connection_status = ConnectionStatus.DISCONNECTED
            self.update_connection_status(connection_status)

    def check_stability(self):
        """Check whether the oven's temperature is stable."""
        MAX_DICONNECTS = 10
        temperature = self.read_temp()
        setpoint = self.get_setpoint()
        if temperature is None or setpoint is None:
            self.disconnected_count += 1
            if self.disconnected_count == MAX_DICONNECTS:
                self.reset_stability()
            elif self.disconnected_count > MAX_DICONNECTS:
                self.disconnected_count = MAX_DICONNECTS + 1  # clamp the value to prevent overflow
        else:  # we are connected and have measurements
            self.disconnected_count = 0
            if self.measurement_is_stable(temperature, setpoint):
                # if we're not already stable, check to see if we can enter the stable state
                if not self.is_stable():
                    self.increment_stability_count()
                    if self.stability_measurement_count >= self.minimum_stability_measurements:
                        self.update_stability_status(True)
            else:
                # we're not stable, update accordingly
                self.reset_stability()

    def update_stability_count(self, count: int):
        """Update the stability measurement count and emit signals on a change."""
        if self.stability_measurement_count != count:
            self.stability_measurement_count = count
            self.stabilityCountChanged.emit(count)

    def increment_stability_count(self):
        """Increment the stability measurement count by 1 and emit signals."""
        self.stability_measurement_count += 1
        self.stabilityCountChanged.emit(self.stability_measurement_count)

    def update_stability_status(self, stable: bool):
        """Update the oven's stability status and emit signals on a change."""
        if self.stable.get() != stable:
            self.stable.set(stable)
            self.stabilityChanged.emit(stable)

    def reset_stability(self):
        """Reset the oven's stability status and counter."""
        self.update_stability_count(0)
        self.update_stability_status(False)

    def is_stable(self) -> bool:
        """Whether the oven's temperature is stable. This is thread safe."""
        return self.stable.get()

    def stability_count(self) -> int:
        """Get the number of stability counts on the oven."""
        return self.stability_measurement_count

    def measurement_is_stable(self, temperature: float, setpoint: float) -> bool:
        """Whether the temperature is within the oven's stability tolerance."""
        return abs(temperature - setpoint) <= self.stability_tolerance

    def set_port(self, port: str):
        """
        Updates the oven's connection port. The oven automatically tries to reconnect. This is
        thread-safe.
        """
        self.port = port
        self.connect()

    def start(self):
        """
        Start the oven's connection timer, which handles maintaining a connection to the physical
        instrument.
        """
        self.moveToThread(self.task_thread)
        self.task_thread.started.connect(self.run)
        self.task_thread.start()

    def run(self):
        connection_timer = Timer(self, 1000, self.connect)  # arbitrary time
        measurement_timer = Timer(
            self, self.measurement_interval, self.read_temp, self.get_setpoint
        )
        stability_timer = Timer(self, self.stability_measurement_interval, self.check_stability)
        for timer in (connection_timer, measurement_timer, stability_timer):
            self.thread().finished.connect(timer.stop)  # type: ignore

        def handle_connection_change(connected: bool):
            if connected:
                connection_timer.stop()
                measurement_timer.start()
            else:
                measurement_timer.stop()
                self.last_temperature = None
                self.last_setpoint = None
                connection_timer.start()

        self.connectionChanged.connect(handle_connection_change)

        connection_timer.start()
        stability_timer.start()

    def stop(self):
        """Stop the oven's automatic monitoring."""
        self.task_thread.quit()
        self.task_thread.wait()


@dataclass
class InstrumentSet:
    """Container for instruments (ovens, etc.)"""

    oven: Union[Oven, "DeveloperOven"]


global INSTRUMENTS
# INSTRUMENTS = InstrumentSet(Oven())



"""Filepaths for sequence settings."""

from .saved import SAVED_SETTINGS

SAVED_SETTINGS_FOLDER = SAVED_SETTINGS.joinpath("sequence")

# files
SEQUENCE_ITEMS_FILE = SAVED_SETTINGS_FOLDER.joinpath("sequence_items_autosave.json")
SEQUENCE_STATE_FILE = SAVED_SETTINGS_FOLDER.joinpath("sequence_state_autosave.json")
OPTIONS_STATE_FILE = SAVED_SETTINGS_FOLDER.joinpath("options_state_autosave.json")
SEQUENCE_DIRECTORY_FILE = SAVED_SETTINGS_FOLDER.joinpath("sequence_directory.json")
NON_EMPTY_DIRECTORY_WARNING_FILE = SAVED_SETTINGS_FOLDER.joinpath(
    "non_empty_directory_warning.json"
)



from fabrial.utility.sequence_builder import PluginCategory

from ...mock_item import MockDataItem


def categories():
    return [
        # simulates someone not using a type checker (please for the love of god use a type checker)
        PluginCategory(
            "Wrong Types", [{"wrong": "type"}, MockDataItem("WrongTypes1", 0)]  # type: ignore
        ),
        PluginCategory(
            "Right Types", [MockDataItem("RightTypes1", 1), MockDataItem("RightTypes2", 2)]
        ),
    ]



from PyQt6.QtWidgets import QLayout, QSizePolicy

from .augmented.widget import Widget


class Container(Widget):
    """
    `Widget` with no contents margins.

    Parameters
    ----------
    layout
        The layout to initialize with.
    horizontal_size_policy
        The horizontal size policy.
    vertical_size_policy
        The vertical size policy.
    """

    def __init__(
        self,
        layout: QLayout,
        horizontal_size_policy: QSizePolicy.Policy = QSizePolicy.Policy.Preferred,
        vertical_size_policy: QSizePolicy.Policy = QSizePolicy.Policy.Preferred,
    ):
        Widget.__init__(self, layout)
        layout.setContentsMargins(0, 0, 0, 0)
        self.setSizePolicy(horizontal_size_policy, vertical_size_policy)


class FixedContainer(Container):
    """Container with a fixed size."""

    def __init__(self, layout: QLayout):
        Container.__init__(self, layout, QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)



from PyQt6.QtCore import Qt
from PyQt6.QtGui import QIcon
from PyQt6.QtWidgets import QLayout

from ..custom_widgets import ParameterDescriptionWidget
from ..utility.descriptions import NO_DESCRIPTION_PROVIDED, DescriptionProvider


class ItemWidget(ParameterDescriptionWidget):
    """
    A widget that holds the data and description for a sequence item.

    Parameters
    ----------
    parameter_widget
        The widget to use for the parameter tab. It can be accessed via `parameter_widget()`.
    title
        The window's initial title. This can be changed via `setWindowTitle()` and accessed via
        `windowTitle()`.
    icon
        The windows initial icon. This can be changed via `setWindowIcon()` and accessed via
        `windowIcon()`.
    description_info
        A `DescriptionInfo` to initialize the description tab with. If this is `None`, the
        description tab will have no description.
    """

    def __init__(
        self,
        parameter_layout: QLayout | None,
        title: str,
        icon: QIcon,
        description_provider: DescriptionProvider | None,
    ):
        ParameterDescriptionWidget.__init__(self, parameter_layout)
        self.setWindowTitle(title)
        self.setWindowIcon(icon)
        description = (
            description_provider.render()
            if description_provider is not None
            else NO_DESCRIPTION_PROVIDED
        )
        self.set_description(description)
        self.windowTitle()
        # make it so you have to close this window to access the application
        self.setWindowModality(Qt.WindowModality.ApplicationModal)

    def show_editable(self, editable: bool):
        """Show the widget. If **editable** is `True`, the parameter tab can be edited."""
        self.parameter_widget().setEnabled(editable)
        self.show()



from ....gamry_integration import GAMRY
from ..base_widget import CategoryWidget


class ElectrochemistryCategoryWidget(CategoryWidget):
    def __init__(self):
        CategoryWidget.__init__(self, "Electrochemistry")

    @staticmethod
    def allowed_to_create():
        return GAMRY.is_valid()



import math
from typing import Any

from comtypes import client  # type: ignore
from impedance_reader import ImpedanceReader

GamryCOM = client.GetModule("C:/Program Files (x86)/Gamry Instruments/Framework/GamryCom.exe")


# --------------------------------------------------------------------------------------------------
# functions
def init_pstat(potentiostat: Any, dc_voltage_vs_reference: float):
    # almost entirely copied from Potentiostatic EIS.exp

    # I turn this on early because it doesn't seem to matter if it's off?
    potentiostat.SetCell(GamryCOM.CellOn)
    potentiostat.SetCtrlMode(GamryCOM.PstatMode)
    potentiostat.SetAchSelect(GamryCOM.GND)
    potentiostat.SetIEStability(GamryCOM.StabilityFast)
    potentiostat.SetCASpeed(3)
    potentiostat.SetSenseSpeedMode(True)
    potentiostat.SetGround(GamryCOM.Float)
    potentiostat.SetIchRange(3.0)
    potentiostat.SetIchRangeMode(False)
    potentiostat.SetIchFilter(2.5)
    potentiostat.SetVchRange(3.0)
    potentiostat.SetVchRangeMode(False)

    potentiostat.SetIchOffsetEnable(True)
    potentiostat.SetVchOffsetEnable(True)

    potentiostat.SetVchFilter(2.5)
    potentiostat.SetAchRange(3.0)
    potentiostat.SetIERangeLowerLimit(0)
    potentiostat.SetIERange(0.03)
    potentiostat.SetIERangeMode(False)
    potentiostat.SetAnalogOut(0.0)
    potentiostat.SetVoltage(dc_voltage_vs_reference)
    potentiostat.SetPosFeedEnable(False)
    potentiostat.SetIruptMode(GamryCOM.IruptOff)


def init_readz(readz: Any, potentiostat: Any, impedance_guess: float):
    # copied from Potentiostatic EIS.exp
    readz.SetSpeed(GamryCOM.ReadZSpeedNorm)
    readz.SetGain(1.0)
    readz.SetINoise(0.0)
    readz.SetVNoise(0.0)
    readz.SetIENoise(0.0)
    readz.SetZmod(impedance_guess)
    readz.SetIdc(potentiostat.MeasureI())


# --------------------------------------------------------------------------------------------------
# values
INITIAL_FREQUENCY = 0.2
FINAL_FREQUENCY = 1000000
POINTS_PER_DECADE = 10
AC_VOLTAGE = 20 / 1000  # 20 mV. I assume that all voltages are in Volts, not millivolts
DC_VOLTAGE = 0
VS_OPEN_CIRCUIT_VOLTAGE = True
IMPEDANCE_GUESS = 1000

LOG_INCREMENT = 1 / POINTS_PER_DECADE
if INITIAL_FREQUENCY > FINAL_FREQUENCY:
    LOG_INCREMENT = -LOG_INCREMENT

MAXIMUM_MEASUREMENTS = math.ceil(
    abs(math.log10(FINAL_FREQUENCY) - math.log10(INITIAL_FREQUENCY)) * POINTS_PER_DECADE
)
# --------------------------------------------------------------------------------------------------
# create devices
device_list = client.CreateObject(GamryCOM.GamryDeviceList)
potentiostat = client.CreateObject(GamryCOM.GamryPC6Pstat)  # should this be GamryPstat instead?
potentiostat.Init(device_list.EnumSections()[0])  # grab first potentiostat
potentiostat.Open()
readz = client.CreateObject(GamryCOM.GamryReadZ)
readz.Init(potentiostat)
init_pstat(potentiostat, DC_VOLTAGE)
init_readz(readz, potentiostat, IMPEDANCE_GUESS)
impedance_reader = ImpedanceReader(
    readz, INITIAL_FREQUENCY, FINAL_FREQUENCY, AC_VOLTAGE, LOG_INCREMENT, MAXIMUM_MEASUREMENTS
)
impedance_reader.run()  # this actually measures and records data

# close and release items
potentiostat.SetCell(GamryCOM.CellOff)
potentiostat.Close()
device_list.Release()

print("Finished")

# TODO: test this script
# TODO: OPEN CIRCUIT VOLTAGE STUFF IF NEEDED



import typing
from typing import Iterable, Mapping, Self

from fabrial import Json, SequenceStep, WidgetDataItem

from .random_step import RandomDataStep

# notice that we import our widget too
from .random_widget import RandomDataWidget

# this constant is used for serializing and deserializing later
DATA_INTERVAL_KEY = "data_interval"


# we inherit from `WidgetDataItem` so Fabrial can use the item
class RandomDataItem(WidgetDataItem):
    """Record random data on an interval; item."""

    def __init__(self, data_interval: float):
        # the only thing we need to do here is create the widget
        self.random_data_widget = RandomDataWidget(data_interval)

    def serialize(self) -> dict[str, Json]:
        # we store the data interval in a dictionary
        return {DATA_INTERVAL_KEY: self.random_data_widget.interval_spinbox.value()}

    @classmethod  # notice the `@classmethod` decorator. You must have this
    def deserialize(cls, serialized_obj: Mapping[str, Json]) -> Self:
        # `cls` is the class, `RandomDataItem` in this case

        # we extract the data interval from the serialized object
        # notice how we use the same dictionary key as in `serialize()`
        data_interval = serialized_obj[DATA_INTERVAL_KEY]
        # `typing.cast()` is for type checkers; it is optional if you
        # aren't type checking
        return cls(typing.cast(float, data_interval))
        # the above line is equivalent to
        # `return RandomDataItem(data_interval)`

    def widget(self) -> RandomDataWidget:
        # this is just a getter for the widget
        return self.random_data_widget

    # the `substeps` parameter is always an empty list for items that don't
    # support subitems. If the item does supports subitems, those items will be
    # converted into sequence steps and passed to this function
    def create_sequence_step(self, substeps: Iterable[SequenceStep]) -> SequenceStep:
        return RandomDataStep(self.random_data_widget.interval_spinbox.value())



from enum import Enum, auto


class SequenceStatus(Enum):
    """Statuses for the sequence."""

    Active = auto()
    Paused = auto()
    Completed = auto()
    Cancelled = auto()
    FatalError = auto()


class SequenceCommand(Enum):
    """Commands that can be send to the sequence thread."""

    Pause = auto()
    Unpause = auto()
    Cancel = auto()
    RaiseFatal = auto()



from types import ModuleType
from typing import Mapping

from PyQt6.QtGui import QCloseEvent
from PyQt6.QtWidgets import QMainWindow, QWidget

from .constants import APP_NAME
from .custom_widgets import TabWidget, YesCancelDialog
from .menu import MenuBar
from .secondary_window import SecondaryWindow
from .tabs import SequenceBuilderTab, SequenceDisplayTab
from .utility import events, images


class MainWindow(QMainWindow):
    def __init__(self, plugin_modules: Mapping[str, ModuleType]):
        self.relaunch = False
        QMainWindow.__init__(self)
        self.setWindowTitle(APP_NAME)
        # create menu bar
        self.menu_bar = MenuBar(self)
        self.setMenuBar(self.menu_bar)
        # self.settings_window = ApplicationSettingsWindow(self) # TODO: fix
        # tabs
        self.sequence_visuals_tab = SequenceDisplayTab()
        self.sequence_tab = SequenceBuilderTab(
            self.sequence_visuals_tab, self.menu_bar.sequence, plugin_modules
        )
        self.tab_widget = TabWidget(
            [
                (self.sequence_tab, "Sequence Builder", images.make_icon("script-block.png")),
                (self.sequence_visuals_tab, "Sequence Visuals", images.make_icon("chart.png")),
            ]
        )
        self.setCentralWidget(self.tab_widget)

        # secondary windows are stored here
        self.secondary_windows: list[QMainWindow] = []

    # ----------------------------------------------------------------------------------------------
    # resizing
    def toggle_fullscreen(self):
        """Toggle fullscreen mode."""
        if self.isFullScreen():
            self.showNormal()
        else:
            self.showFullScreen()

    def shrink(self):
        """Shrink the window to its minimum size. Exits fullscreen mode."""
        if self.isFullScreen():
            self.showNormal()
        self.resize(self.sizeHint())

    # ----------------------------------------------------------------------------------------------
    # multiple windows
    def new_window(self, title: str, central_widget: QWidget) -> SecondaryWindow:
        """
        Create a new window owned by the main window. The window is automatically shown.

        Parameters
        ----------
        title
            The window title.
        central_widget
            The widget to show inside the secondary window.

        Returns
        -------
        The created window.
        """
        window = SecondaryWindow(title, central_widget)
        self.secondary_windows.append(window)
        window.closed.connect(lambda: self.secondary_windows.remove(window))
        window.show()
        return window

    # ----------------------------------------------------------------------------------------------
    # closing the window
    def closeEvent(self, event: QCloseEvent | None):  # overridden method
        """Prevent the window from closing if a sequence or stability check are running."""
        if event is not None:
            if self.allowed_to_close():
                self.save_on_close()
                event.accept()
            else:
                event.ignore()

    def allowed_to_close(self) -> bool:
        """Determine if the window should close."""
        # only close if a sequence is not running
        return True  # TODO
        process_widget = self.sequence_tab.sequence_tree
        if process_widget.is_running():
            if YesCancelDialog(
                "Are you sure you want to exit?", "A sequence is currently running."
            ).run():
                process_widget.cancelCommand.emit()
                while process_widget.is_running():
                    events.PROCESS_EVENTS()
            else:
                return False
        return True

    def save_on_close(self):
        """Save all data that gets saved on closing. Call this when closing the application."""
        self.sequence_tab.save_on_close()

    def should_relaunch(self) -> bool:
        """Whether the application should relaunch."""
        return self.relaunch

    def set_relaunch(self, should_relaunch: bool):
        """Set whether the application should relaunch."""
        self.relaunch = should_relaunch

    # ----------------------------------------------------------------------------------------------
    # settings
    def show_settings(self):
        """Show the application settings. These settings are saved when the window closes."""
        return
        # self.settings_window.show()



from fabrial.utility.sequence_builder import PluginCategory

from ...mock_item import MockDataItem


def categories() -> list[PluginCategory]:
    # intentionally out of order
    return [
        PluginCategory("Normal2", [MockDataItem("Normal22", 22), MockDataItem("Normal21", 21)]),
        PluginCategory("Normal1", [MockDataItem("Normal11", 11), MockDataItem("Normal12", 12)]),
    ]



import csv
import math
import time
from typing import Any

from comtypes import client  # type: ignore


class ImpedanceReader:
    def __init__(
        self,
        readz: Any,
        initial_frequency: float,
        final_frequency: float,
        ac_voltage: float,
        log_increment: float,
        maximum_measurements: int,
    ):
        self.readz = readz
        self.initial_frequency = initial_frequency
        self.final_frequency = final_frequency
        self.ac_voltage = ac_voltage
        self.log_increment = log_increment
        self.maximum_measurements = maximum_measurements

        self.measurement_count = 0
        self.calibration_measurement = True
        self.start_time = 0.0
        self.connection = client.GetEvents(self.readz, self)

        # create file and write headers
        self.file = open(
            "output.DTA",
            "w",
            1,  # line buffer
            newline="",  # necessary for CSVwriter
        )
        self.writer = csv.writer(self.file, delimiter="\t")
        self.writer.writerows(
            [
                ["EXPLAIN"],
                ["TAG", "EISPOT"],
                ["TITLE", "LABEL", "Potentiostatic EIS", "Test &Identifier"],
                [],
                ["ZCURVE", "TABLE"],
                [
                    "",
                    "Pt",
                    "Time",
                    "Freq",
                    "Zreal",
                    "Zimag",
                    "Zsig",
                    "Zmod",
                    "Zphz",
                    "Idc",
                    "Vdc",
                    "IERange",
                ],
                ["", "#", "s", "Hz", "ohm", "ohm", "V", "ohm", "°", "A", "V", "#"],
            ],
        )

    def run(self):
        self.start_time = time.time()
        # calibration measurement (triggers a chain of other measurements)
        self.readz.Measure(self.initial_frequency, self.ac_voltage)
        while not self.is_finished():
            client.PumpEvents(0.1)

        self.file.close()
        self.connection.disconnect()
        del self.connection

    def is_finished(self) -> bool:
        return self.measurement_count > self.maximum_measurements

    def record_measurement(self):
        self.writer.writerow(
            [
                "",  # for initial tab
                self.measurement_count,
            ]
            + [  # round all values to 6 decimal places
                round(value, 6)
                for value in [
                    time.time() - self.start_time,
                    self.frequency(),
                    self.real_impedance(),
                    self.imaginary_impedance(),
                    self.impedance_standard_deviation(),
                    self.impedance_magnitude(),
                    self.impedance_phase(),
                    self.dc_current(),
                    self.dc_voltage(),
                    self.ie_range(),
                ]
            ]
        )

    # ----------------------------------------------------------------------------------------------
    # COM functions
    def _IGamryReadZEvents_OnDataDone(self, this: Any, error_status: int):
        print(self.measurement_count)
        if error_status != 0:
            # in general ignoring the error status because it isn't relevant for this example
            print("An error occurred")

        # redo the first measurement once so the potentiostat is calibrated
        if self.calibration_measurement:
            self.calibration_measurement = False
            self.readz.Measure(self.initial_frequency, self.ac_voltage)
            return

        # if we got here we are actually measuring and recording data
        self.record_measurement()
        self.measurement_count += 1
        if not self.is_finished():
            desired_frequency = math.pow(
                10,
                math.log10(self.initial_frequency) + self.measurement_count * self.log_increment,
            )
            self.readz.Measure(desired_frequency, self.ac_voltage)
            return

    def _IGamryReadZEvents_OnDataAvailable(self, this: Any):
        # unused
        pass

    # ----------------------------------------------------------------------------------------------
    # measurement values
    def frequency(self) -> float:
        return self.readz.Zfreq()

    def real_impedance(self) -> float:
        return self.readz.Zreal()

    def imaginary_impedance(self) -> float:
        return self.readz.Zimag()

    def impedance_standard_deviation(self) -> float:
        return self.readz.Zsig()

    def impedance_magnitude(self) -> float:
        return self.readz.Zmod()

    def impedance_phase(self) -> float:
        return self.readz.Zphz()

    def dc_current(self) -> float:
        return self.readz.Idc()

    def dc_voltage(self) -> float:
        return self.readz.Vdc()

    def ie_range(self) -> float:
        return self.readz.IERange()



from PyQt6.QtWidgets import QBoxLayout, QFormLayout, QLayout, QLayoutItem, QStackedLayout, QWidget


def add_sublayout[LayoutType: QLayout](
    outer_layout: QBoxLayout, inner_layout: LayoutType
) -> LayoutType:
    """
    Add an inner layout to an outer layout.

    Parameters
    ----------
    outer_layout
        The outer layout to add an inner layout to.
    inner_layout
        The layout to add.

    Returns
    -------
    The **inner_layout**.
    """
    outer_layout.addLayout(inner_layout)
    return inner_layout


def add_to_layout(layout: QBoxLayout | QStackedLayout, *widgets: QWidget):
    """
    Adds widgets to a QBoxLayout or QStackedLayout. Cannot handle alignment or stretch.

    Parameters
    ----------
    layout
        The layout to add the widgets to.
    widgets
        The widgets to add.
    """
    for widget in widgets:
        layout.addWidget(widget)


def add_to_form_layout(layout: QFormLayout, *item_pair: tuple[QWidget | str, QWidget]):
    """
    Adds widgets to a QFormLayout.

    Parameters
    ----------
    layout
        The layout to add widgets to.
    item_pair
        Item pair(s) in the form of (LEFT_ITEM, RIGHT_ITEM), where LEFT_ITEM is the item on the left
        and RIGHT_ITEM is the item on the right. If LEFT_ITEM is a `str`, a label will be created
        with LEFT_ITEM as the text.
    """
    for pair in item_pair:
        layout.addRow(pair[0], pair[1])


def clear_layout(layout: QLayout):
    """Removes all widgets and sublayouts from the layout."""
    for i in reversed(range(layout.count())):
        # setting a widget's parent to None schedules it for deletion
        layout_item: QLayoutItem = layout.takeAt(i)  # type: ignore
        widget = layout_item.widget()
        if widget is not None:  # the item is a widget
            widget.setParent(None)
        else:  # the item is a layout
            contained_layout = layout_item.layout()
            if contained_layout is not None:
                contained_layout.setParent(None)



from .data_item import DataItem, WidgetDataItem
from .item_widget import ItemWidget
from .tree_items import CategoryItem, SequenceItem, TreeItem
from .tree_models import OptionsModel, SequenceModel
from .tree_views import OptionsTreeWidget, SequenceTreeWidget



from fabrial.utility.sequence_builder import PluginCategory

from ...mock_item import MockDataItem


def categories() -> list[PluginCategory]:
    # intentionally out of order
    return [
        PluginCategory(
            "Outer Category",
            [MockDataItem("Category2Outer", 2)],
            subcategories=[
                PluginCategory(
                    "Nested Category",
                    [MockDataItem("Category2Item2", 22), MockDataItem("Category2Item1", 21)],
                )
            ],
        ),
    ]



import json
from abc import abstractmethod
from typing import Any, Iterable

from PyQt6.QtCore import (
    QAbstractItemModel,
    QByteArray,
    QDataStream,
    QIODevice,
    QMimeData,
    QModelIndex,
    Qt,
)
from PyQt6.QtWidgets import QApplication

from ...classes import QABC
from ..tree_items import RootItem, TreeItem

JSON = "application/json"


class TreeModel[Item: TreeItem](QAbstractItemModel, QABC):
    """
    `QAbstractItemModel` for representing trees.

    Parameters
    ----------
    name
        The name displayed at the top of the widget.
    items
        The items to initialize with. Can be empty.
    """

    def __init__(self, title: str, items: Iterable[Item]):
        QAbstractItemModel.__init__(self)

        self.name = title
        self.root_item: RootItem[Item] = RootItem()
        self.root_item.append_subitems(items)

        self.base_flag = Qt.ItemFlag.ItemIsEnabled

    def title(self) -> str:
        """Get the model's title."""
        return self.name

    def root(self) -> RootItem[Item]:
        """Get the root item."""
        return self.root_item

    def copy_items(self, indexes: Iterable[QModelIndex]):
        """Copy items to the clipboard."""
        data = self.mimeData(sorted(indexes))
        clipboard = QApplication.clipboard()
        if clipboard is not None:
            clipboard.setMimeData(data)

    def is_enabled(self) -> bool:
        "Whether the model's items are enabled."
        return self.base_flag != Qt.ItemFlag.NoItemFlags

    def set_enabled(self, enabled: bool):
        """Set whether the model's items are enabled."""
        self.base_flag = Qt.ItemFlag.ItemIsEnabled if enabled else Qt.ItemFlag.NoItemFlags

    def flags(self, index: QModelIndex) -> Qt.ItemFlag:  # overridden
        flags = self.base_flag
        item = self.get_item(index)
        if item is None:
            if self.root_item.supports_subitems():
                flags |= Qt.ItemFlag.ItemIsDropEnabled
            return flags

        if item.supports_subitems():
            flags |= Qt.ItemFlag.ItemIsDropEnabled
        if item.supports_dragging():
            flags |= Qt.ItemFlag.ItemIsDragEnabled | Qt.ItemFlag.ItemIsSelectable
        return flags

    @abstractmethod
    def data(self, index: QModelIndex, role: int | None = None) -> Any:  # overridden
        ...

    def get_item(self, index: QModelIndex) -> Item | None:
        """Get the item at the provided **index**. Returns `None` if **index** is invalid."""
        if index.isValid():
            # this uses C++ witchcraft to get the item at the index
            # look up the docs for QModelIndex
            # I think it is related to the index() function
            # it's something with pointers, idk
            item: Item | None = index.internalPointer()
            if item is not None:
                return item
        return None

    # ----------------------------------------------------------------------------------------------
    # overridden
    def parent(self, index: QModelIndex) -> QModelIndex:  # type: ignore
        item = self.get_item(index)
        if item is None:
            return QModelIndex()

        parent_item = item.parent()
        if parent_item is None:
            return QModelIndex()

        row = parent_item.index_in_parent()
        if row is None:
            return QModelIndex()

        return self.createIndex(row, 0, parent_item)

    def columnCount(self, parent: QModelIndex | None = None) -> int:  # overridden
        return 1

    def rowCount(self, index: QModelIndex = QModelIndex()) -> int:  # overridden
        item: TreeItem = self.get_item(index) or self.root_item
        return item.subitem_count()

    # overridden
    def headerData(
        self, section: int, orientation: Qt.Orientation, role: int | None = None
    ) -> str | None:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self.title()
        return None

    # overridden
    def index(
        self, row: int, column: int, parent_index: QModelIndex = QModelIndex()
    ) -> QModelIndex:
        parent_item: TreeItem = self.get_item(parent_index) or self.root_item
        item = parent_item.get_subitem(row)
        if item is None:
            return QModelIndex()
        return self.createIndex(row, column, item)

    def mimeData(self, indexes: Iterable[QModelIndex]) -> QMimeData | None:  # overridden
        mime_data = QMimeData()
        encoded_data = QByteArray()
        stream = QDataStream(encoded_data, QIODevice.OpenModeFlag.WriteOnly)
        for index in indexes:
            item = self.get_item(index)
            if item is not None:
                text = json.dumps(item.serialize())
                stream.writeQString(text)

        if encoded_data.length() == 0:
            return None  # return `None` instead of an empty `QMimeData`

        mime_data.setData(JSON, encoded_data)
        return mime_data

    def mimeTypes(self) -> list[str]:  # overridden
        return [JSON]

    def expand_event(self, index: QModelIndex):
        """Handle an item being expanded in the view."""
        item = self.get_item(index)
        if item is not None:
            item.expand_event()

    def collapse_event(self, index: QModelIndex):
        """Handle an item being collapsed in the view."""
        item = self.get_item(index)
        if item is not None:
            item.collapse_event()



from __future__ import annotations

from typing import TYPE_CHECKING

from PyQt6.QtCore import Qt
from PyQt6.QtGui import QCloseEvent
from PyQt6.QtWidgets import QApplication, QTabWidget, QVBoxLayout

from ...constants import APP_NAME

if TYPE_CHECKING:
    from ...main_window import MainWindow

from ...tabs import OvenControlTab, SequenceBuilderTab
from ...utility import images
from ..augmented import BiggerButton
from ..container import Container
from .gamry import GamrySettingsTab
from .oven import OvenSettingsTab
from .sequence import SequenceSettingsTab


class ApplicationSettingsWindow(Container):
    """Settings window the application's settings."""

    def __init__(self, main_window: MainWindow) -> None:
        self.main_window = main_window
        layout = QVBoxLayout()
        Container.__init__(self, layout)
        self.setWindowTitle(f"{APP_NAME} - Settings")

        tab_widget = QTabWidget(self)
        self.sequence_tab = SequenceSettingsTab()
        self.oven_tab = OvenSettingsTab()
        self.gamry_tab = GamrySettingsTab()
        tab_widget.addTab(
            self.sequence_tab, images.make_icon(SequenceBuilderTab.ICON_FILE), "Sequence"
        )
        tab_widget.addTab(self.oven_tab, images.make_icon(OvenControlTab.ICON_FILE), "Oven")
        tab_widget.addTab(self.gamry_tab, images.make_icon("lightning.png"), "Gamry")
        layout.addWidget(tab_widget)

        layout.addWidget(
            BiggerButton("Relaunch", self.handle_relaunch_request, size_scalars=(2, 2)),
            alignment=Qt.AlignmentFlag.AlignCenter,
        )

    def handle_relaunch_request(self):
        """Runs when the relaunch button is pressed."""
        self.main_window.set_relaunch(True)
        QApplication.quit()

    def closeEvent(self, event: QCloseEvent | None):  # overridden
        if event is not None:
            for tab in (self.sequence_tab, self.oven_tab, self.gamry_tab):
                tab.save_on_close()
        Container.closeEvent(self, event)



from abc import ABCMeta
from typing import Protocol

from PyQt6.QtCore import QObject


class QProtocolMeta(type(QObject), type(Protocol)):  # type: ignore
    """Metaclass combining `Protocol` and `QObject`."""


class QProtocol(Protocol, metaclass=QProtocolMeta):
    pass


class ABCQObjectMeta(ABCMeta, type(QObject)):  # type: ignore
    pass


class QABCMeta(ABCMeta, type(QObject)):  # type: ignore
    """Metaclass combing `ABCMeta` and QObject's metaclass."""


class QABC(metaclass=QABCMeta):
    """`ABC` with support for `QObject`s."""



import csv
import math
import os
import time
from contextlib import ExitStack
from dataclasses import dataclass
from io import TextIOWrapper
from typing import Any, Iterable

from PyQt6.QtWidgets import QMessageBox

from .....classes import AbstractGraphingProcess, ProcessRunner
from .....gamry_integration import GAMRY, ImpedanceReader, Potentiostat
from .....utility import dataframe
from . import encoding
from .encoding import FileFormat, Headers

try:
    from comtypes import client  # type: ignore
except ImportError:
    pass


@dataclass
class ImpedanceReaderData:
    file: TextIOWrapper
    plot_index: int
    frequency: float
    measurement_count: int = 0
    calibration_measurement: bool = True
    retry_count: int = 0


class EISProcess(AbstractGraphingProcess):
    """Run a clone of the Gamry Potentiostatic EIS.exp file."""

    def __init__(self, runner: ProcessRunner, data: dict[str, Any], name: str):
        AbstractGraphingProcess.__init__(self, runner, data, name)
        # unpack the dictionary values for easier access
        self.initial_frequency: float = data[encoding.INITIAL_FREQUENCY]
        self.final_frequency: float = data[encoding.FINAL_FREQUENCY]
        self.points_per_decade: int = data[encoding.POINTS_PER_DECADE]
        self.ac_voltage: float = data[encoding.AC_VOLTAGE] / 1000  # convert from mV to V
        self.dc_voltage: float = data[encoding.DC_VOLTAGE]
        self.vs_open_circuit_voltage = data[encoding.DC_VOLTAGE_REFERENCE] == encoding.VS_EOC
        self.impedance_guess: float = data[encoding.ESTIMATED_IMPEDANCE]
        self.impedance_reader_speed_option: str = data[encoding.IMPEDANCE_READER_SPEED]
        self.pstat_identifiers: list[str] = data[encoding.SELECTED_PSTATS]

        self.log_increment = 1 / self.points_per_decade
        if self.initial_frequency > self.final_frequency:
            # if we are sweeping high frequency to low frequency, we need to decrease in frequency
            self.log_increment = -self.log_increment

        self.maximum_measurements = math.ceil(
            abs(math.log10(self.final_frequency) - math.log10(self.initial_frequency))
            * self.points_per_decade
        )

        self.impedance_reader_data: dict[ImpedanceReader, ImpedanceReaderData] = dict()

    def run(self) -> None:
        # open all files, potentiostats, and impedance readers using a context manager so they
        # get closed automatically
        with ExitStack() as context_manager:
            impedance_readers = self.create_impedance_readers(context_manager)
            try:
                # do the first measurement (this starts a chain)
                for impedance_reader in impedance_readers:
                    impedance_reader.measure(self.initial_frequency, self.ac_voltage)
                # run while at least one impedance reader still has measurements to do or until we
                # get #canceled
                while not self.is_canceled():
                    # this causes a bunch of signals to fire which do most of the data acquisition
                    # and recording
                    client.PumpEvents(0.1)
                    self.wait(50)
                    for impedance_reader in impedance_readers:
                        if not self.is_finished(impedance_reader):
                            break
                    else:  # the for loop finished naturally
                        break
            except Exception:
                pass
            finally:
                self.post_run(impedance_readers)

    def post_run(self, impedance_readers: Iterable[ImpedanceReader]):
        """Post-run tasks."""
        # save the bode plots for each potentiostat
        for impedance_reader in impedance_readers:
            impedance_reader_data = self.impedance_reader_data[impedance_reader]
            identifier = impedance_reader.potentiostat().identifier()
            self.graphing_signals().saveFig.emit(
                impedance_reader_data.plot_index,
                os.path.join(self.directory(), f"{identifier}.png"),
            )

    @staticmethod
    def directory_name():
        return "Electrochemical Impedance Spectroscopy"

    def handle_data_ready(self, impedance_reader: ImpedanceReader, success: bool):
        """
        Handle data being ready for an impedance reader. The first measurement is run twice; the
        first time calibrates the potentiostat and the second time actually records data.
        """
        # unless the impedance reader is finished with its frequency sweep, each call to this
        # function will queue another call for the same impedance reader
        if self.is_canceled():
            return

        impedance_reader_data = self.impedance_reader_data[impedance_reader]

        if impedance_reader_data.calibration_measurement:  # if this is the calibration measurement
            impedance_reader_data.calibration_measurement = False
            impedance_reader.measure(impedance_reader_data.frequency, self.ac_voltage)
            return

        if not success:
            # if we failed to read, retry 10 times
            if impedance_reader_data.retry_count < 10:
                impedance_reader.measure(impedance_reader_data.frequency, self.ac_voltage)
                impedance_reader_data.retry_count += 1
                return
            # if we still failed, ask the user what to do
            identifer = impedance_reader.potentiostat().identifier()
            self.send_message(
                f"Failed to take measurement for potentiostat {identifer} "
                f"at frequency {impedance_reader_data.frequency}. "
                "Continue to next frequency, retry measurement, "
                "or abort experiment?",
                QMessageBox.StandardButton.Yes
                | QMessageBox.StandardButton.Retry
                | QMessageBox.StandardButton.Abort,
                {QMessageBox.StandardButton.Yes: "Next"},
            )
            response = self.wait_on_response()
            if response is None:  # the process has been canceled
                return
            match response:
                case QMessageBox.StandardButton.Yes:  # next measurement
                    pass
                case QMessageBox.StandardButton.Retry:  # retry measurement
                    impedance_reader_data.retry_count = 0
                    impedance_reader.measure(impedance_reader_data.frequency, self.ac_voltage)
                    return
                case QMessageBox.StandardButton.Abort:  # abort experiment
                    self.cancel()
                    return

        # if we are here the measurement either succeeded or was skipped (so move to next frequency)
        self.record_measurement(impedance_reader_data.measurement_count, impedance_reader)

        impedance_reader_data.measurement_count += 1
        if not self.is_finished(impedance_reader):
            desired_frequency = math.pow(
                10,
                math.log10(self.initial_frequency)
                + impedance_reader_data.measurement_count * self.log_increment,
            )
            impedance_reader_data.frequency = desired_frequency
            impedance_reader.measure(impedance_reader_data.frequency, self.ac_voltage)

    def is_finished(self, impedance_reader: ImpedanceReader) -> bool:
        """Whether the impedance reader has finished its frequency sweep."""
        measurement_count = self.impedance_reader_data[impedance_reader].measurement_count
        if measurement_count > self.maximum_measurements:
            return True
        return False

    def create_impedance_readers(self, context_manager: ExitStack) -> list[ImpedanceReader]:
        """
        Create, initialize, and return the process' impedance readers.

        Parameters
        ----------
        context_manager
            The context manager used to automatically close and cleanup the devices.
        """
        impedance_readers: list[ImpedanceReader] = []
        for index, identifer in enumerate(self.pstat_identifiers):
            # make the potentiostat
            potentiostat = context_manager.enter_context(
                Potentiostat(GAMRY.com_interface(), identifer)
            )
            # measure the open circuit voltage to calculate the voltage vs. Eref
            if self.vs_open_circuit_voltage:
                open_circuit_voltage = potentiostat.measure_open_circuit_voltage()
                voltage_vs_reference = self.dc_voltage + open_circuit_voltage
            else:
                voltage_vs_reference = self.dc_voltage
            potentiostat.initialize(voltage_vs_reference)
            # make the impedance reader
            impedance_reader = context_manager.enter_context(
                ImpedanceReader(potentiostat)
            ).initialize(
                self.impedance_guess,
                encoding.IMPEDANCE_READER_SPEED_DICT[self.impedance_reader_speed_option],
            )
            impedance_readers.append(impedance_reader)
            # connect the impedance reader's signals
            # this is weirdly written because non-primitive types are passed by pointer for signals
            impedance_reader.dataReady.connect(
                lambda success, reader=impedance_reader: self.handle_data_ready(reader, success)
            )
            # make the file for the reader
            file = self.create_file(impedance_reader, context_manager)
            # initialize the plot for the reader
            self.init_plot(impedance_reader, index)
            # create a new entry in the impedance reader bookkeeping dictionary
            self.impedance_reader_data[impedance_reader] = ImpedanceReaderData(
                file, index, self.initial_frequency
            )

        return impedance_readers

    def create_file(
        self, impedance_reader: ImpedanceReader, context_manager: ExitStack
    ) -> TextIOWrapper:
        """
        Create the data file for an impedance reader and their write the header.

        Parameters
        ----------
        reader
            The impedance reader to create the file for.
        context_manager
            The context manager used to automatically close the file.
        """
        identifier = impedance_reader.potentiostat().identifier()
        # file names are based on the potentiostat identifier
        file = context_manager.enter_context(
            open(
                os.path.join(self.directory(), f"{identifier}.DTA"),
                "w",
                1,
                newline="",
            )
        )
        writer = csv.writer(file, delimiter=FileFormat.DELIMETER)
        writer.writerows(Headers.GAMRY_HEADERS)
        writer.writerows(Headers.DATA_HEADERS)

        return file

    def init_plot(self, impedance_reader: ImpedanceReader, index: int):
        """Initialize the plot for an impedance reader."""
        identifier = impedance_reader.potentiostat().identifier()
        self.init_scatter_plot(
            index,
            identifier,
            f"Bode Plot for {identifier}",
            "Frequency (Hz)",
            "Impedance Magnitude (kΩ)",
            "Z-Curve",
            symbol_color="lightskyblue",
        )
        self.graphing_signals().setLogScale.emit(index, True, False)

    def record_measurement(self, measurement_count: int, reader: ImpedanceReader):
        """
        Record a measurement in the data file and on the graph. Only the first potentiostat's
        measurement is shown on the graph.

        Parameters
        ----------
        measurement_count
            The current measurement count.
        reader
            The `ImpedanceReader` to get data from.
        """
        impedance_reader_data = self.impedance_reader_data[reader]

        frequency = reader.frequency()
        impedance_magnitude = reader.impedance_magnitude()
        csv.writer(impedance_reader_data.file, delimiter=FileFormat.DELIMETER).writerow(
            [
                "",  # for initial tab
                measurement_count,
            ]
            + [
                round(value, 6)
                for value in [
                    time.time() - self.start_time(),
                    frequency,
                    reader.real_impedance(),
                    reader.imaginary_impedance(),
                    reader.impedance_standard_deviation(),
                    impedance_magnitude,
                    reader.impedance_phase(),
                    reader.dc_current(),
                    reader.dc_voltage(),
                    reader.ie_range(),
                ]
            ]
        )
        self.graphing_signals().addPoint.emit(
            impedance_reader_data.plot_index, frequency, impedance_magnitude / 1000  # convert to kΩ
        )

    def metadata(self):  # overridden
        return dataframe.add_to_dataframe(
            AbstractGraphingProcess.metadata(self),
            {
                encoding.SELECTED_PSTATS: " ".join(self.pstat_identifiers),  # space separate them
                encoding.DC_VOLTAGE: self.dc_voltage,
                encoding.DC_VOLTAGE_REFERENCE: self.data()[encoding.DC_VOLTAGE_REFERENCE],
                encoding.INITIAL_FREQUENCY: self.initial_frequency,
                encoding.FINAL_FREQUENCY: self.final_frequency,
                encoding.POINTS_PER_DECADE: self.points_per_decade,
                encoding.AC_VOLTAGE: self.data()[encoding.AC_VOLTAGE],  # mV
                encoding.ESTIMATED_IMPEDANCE: self.impedance_guess,
                encoding.IMPEDANCE_READER_SPEED: self.impedance_reader_speed_option,
            },
        )



"""
Utilities for serializing and deserializing objects. When deserializing, the object type is read
from the file.
"""

from __future__ import annotations

import json
import tomllib
import typing
from abc import abstractmethod
from os import PathLike
from typing import Any, Mapping, Protocol, Self, Sequence

TYPE = "type"
DESERIALIZABLE_CLASSES: dict[str, type[Deserialize]] = {}

type Json = Mapping[str, Json] | Sequence[Json] | str | int | float | bool | None


class Deserialize(Protocol):
    """An object that can be deserialized without knowing the type ahead of time."""

    @classmethod
    @abstractmethod
    def deserialize(cls, serialized_obj: Mapping[str, Json]) -> Self:
        """Build the object from a JSON-like structure."""
        ...

    def __init_subclass__(cls):  # runs when `Deserialize` is subclasses
        typename = str(cls).split("'")[1]
        # ^
        # for example, str(int) gives "<class 'int'>", so we split on "'" to get
        # ["<class ", "int", ">"] and take the "int" part
        DESERIALIZABLE_CLASSES[typename] = cls  # store the type


class Serialize(Protocol):
    """An object that can be serialized into a type-tagged JSON dictionary."""

    @abstractmethod
    def serialize(self) -> dict[str, Json]:
        """Convert the object to a JSON-like structure. This method does not add a type tag."""
        ...

    def serialize_tagged(self) -> dict[str, Json]:
        """Serialize and add a type tag. This calls `serialize()`."""
        # str(TYPE) gives "<class 'TYPE'>" so we just extract the TYPE part
        type_tagged_dict: dict[str, Json] = {TYPE: str(type(self)).split("'")[1]}
        type_tagged_dict.update(self.serialize())
        return type_tagged_dict

    def save_json(self, file: PathLike[str] | str) -> bool:
        """Save the object to a JSON file. Returns whether the operation succeeded."""
        try:
            with open(file, "w") as f:
                json.dump(self.serialize_tagged(), f)
            return True
        except Exception:
            return False


def get_type(typename: str) -> type[Deserialize]:
    """Get a `Deserialize` subtype from a string representation."""
    try:
        return DESERIALIZABLE_CLASSES[typename]  # defined at the top of the file
    except KeyError:
        raise KeyError(f"`{typename}` does not represent a `Deserialize` object")


def deserialize(serialized_obj: Json) -> Any:
    """
    Deserialize **serialized_obj**. Any type-tagged dictionaries are deserialized into the
    corresponding `Deserialize` object.
    """

    # deserializes a dictionary
    def inner_deserialize_dict(item: Mapping[str, Json]) -> Any:
        try:
            # get the typename key
            typename = typing.cast(str, item[TYPE])  # cast assumes `typename` is `str`
        except KeyError:
            # if we got here the `item` dictionary is not a `Deserialize` object
            return {  # recurse into the rest of the dictionary and deserialize it
                inner_key: inner_deserialize_json(inner_item)
                for inner_key, inner_item in item.items()
            }
        # if we get here we have a type tag
        cls = get_type(typename)  # get the actual type
        # convert the dictionary into the actual `Deserialize` object
        return cls.deserialize(item)

    # deserializes a list
    def inner_deserialize_list(items: Sequence[Json]) -> Any:
        deserialized_items = []
        for inner_item in items:
            # replace each element in the list with its deserialized version
            deserialized_items.append(inner_deserialize_json(inner_item))
        # return the modified list
        return deserialized_items

    # helper function to recurse into JSON structures
    def inner_deserialize_json(item: Json) -> Any:
        # if the item is a dictionary, deserialize the dictionary
        if isinstance(item, dict):
            return inner_deserialize_dict(item)
        # if the item is a list, deserialize the list
        elif isinstance(item, list):
            return inner_deserialize_list(item)
        # otherwise the item is primitive, don't deserialize
        else:
            return item

    return inner_deserialize_json(serialized_obj)


def load_json(file: PathLike[str] | str) -> Any:
    """
    Load an object from a JSON file. Any type-tagged dictionaries are deserialized into the
    corresponding `Deserialize` object.
    """
    with open(file, "r") as f:
        object_as_dict = json.load(f)
    return deserialize(object_as_dict)


def load_toml(file: PathLike[str] | str) -> Any:
    """
    Load an object from a TOML file. Any type-tagged dictionaries are deserialized into the
    corresponding `Deserialize` object.
    """
    with open(file, "rb") as f:
        object_as_dict = tomllib.load(f)
    return deserialize(object_as_dict)



import logging
import sys
from types import TracebackType

from PyQt6 import QtCore
from PyQt6.QtCore import QMessageLogContext, QtMsgType

from ..constants import APP_NAME
from ..custom_widgets import OkDialog
from . import events


def exception_handler(
    exception_type: type[BaseException], exception: BaseException, trace: TracebackType | None
):
    """
    This replaces `sys.excepthook`. It logs uncaught exceptions, then runs `sys.__excepthook__()`.
    """
    if not issubclass(exception_type, KeyboardInterrupt):  # don't notify/log `KeyboardInterrupt`
        # log the exception
        logging.getLogger(__name__).critical(
            "Uncaught exception", exc_info=(exception_type, exception, trace)
        )
        show_error(  # notify the user
            "Fatal Application Error",
            f"{APP_NAME} encountered a fatal error. "
            "Please save the contents of the error log and report the issue.\n\n"
            f"{APP_NAME} will now exit.",
        )
    sys.exit()  # exit


def suppress_warnings():
    """Suppress unnecessary warnings from PyQt6 (version 6.9.1 has a bug on Windows)."""

    def warning_suppressor(msg_type: QtMsgType, context: QMessageLogContext, msg: str | None):
        match msg_type:
            case QtMsgType.QtWarningMsg:
                return
            case _:
                print(msg)

    QtCore.qInstallMessageHandler(warning_suppressor)


def show_error(title: str, message: str):
    """Show an error to the user (this is just an `OkDialog`)."""
    OkDialog(title, message).exec()


def show_error_delayed(title: str, message: str):
    """Wait until the event loop is running, then show an error to the user."""
    events.delay_until_running(lambda: show_error(title, message))


def set_up_logging():
    """Configure the root logger of the `logging` module."""
    logging.basicConfig(
        level=logging.INFO,  # log INFO and up
        style="{",  # used for the format specifier
        datefmt="%Y-%m-%d %H:%M:%S",  # datetime format
        # show datetime, error level, logger name, the calling function's name, and the log message
        format="{asctime}.{msecs:.0f} {levelname} {message} - {name}:{funcName}()",
        filename="lastrun.log",  # send to `lastrun.log`
        filemode="w",  # wipe the file between runs
    )



from .classes import LineSettings, PlotSettings, SequenceStep, StepRunner
from .main_window import MainWindow
from .sequence_builder import DataItem, ItemWidget, WidgetDataItem
from .utility.descriptions import FilesDescription, Substitutions, TextDescription
from .utility.sequence_builder import PluginCategory
from .utility.serde import Deserialize, Json, Serialize, deserialize, load_json, load_toml



from typing import Iterable, Mapping, Self

from PyQt6.QtGui import QIcon

from fabrial import DataItem, Json, SequenceStep


class MockDataItem(DataItem):
    """A `DataItem` implementation for testing."""

    def __init__(self, name: str, number: int):
        self.name = name
        self.number = number

    @classmethod
    def deserialize(cls, serialized_obj: Mapping[str, Json]) -> Self:  # implementation
        raise NotImplementedError("This shouldn't have been tested")

    def serialize(self) -> dict[str, Json]:  # implementation
        raise NotImplementedError("This shouldn't have been tested")

    def display_name(self) -> str:  # implementation
        return self.name

    def icon(self) -> QIcon:  # implementation
        raise NotImplementedError("This shouldn't have been tested")

    def open_event(self, editable: bool):  # implementation
        raise NotImplementedError("This shouldn't have been tested")

    def create_sequence_step(self, substeps: Iterable[SequenceStep]) -> SequenceStep:
        raise NotImplementedError("`create_sequence_step()` called on `MockDataItem`")

    # testing
    def __eq__(self, other: Self):  # type: ignore
        return self.name == other.name and self.number == other.number

    # debugging
    def __repr__(self) -> str:
        return f"{self.__class__.__name__} {{ name: {self.name!r}, number: {self.number!r} }}"



import os

from PyQt6.QtGui import QIcon, QPixmap

from ..constants.paths.icons import INTERNAL_ICONS


def get_image_file(icon_filename: str) -> str:
    """Get the full filepath to an icon in the application's internal icon folder."""
    return os.path.join(INTERNAL_ICONS, icon_filename)


def make_icon(icon_filename: str) -> QIcon:
    """
    Create a QIcon from an icon file in the application's internal icons folder.

    Parameters
    ----------
    icon_filename
        The icon filename inside the icons folder WITH the file extension.

    Returns
    -------
    A `QIcon` of the icon file.
    """
    return QIcon(get_image_file(icon_filename))


def make_pixmap(image_filename: str) -> QPixmap:
    """
    Create a QPixmap from an image file in the application's internal icons folder.

    Parameters
    ----------
    image_filename
        The image filename inside the icons folder WITH the file extension.
    Returns
    -------
    A `QPixmap` of the image file.
    """
    return QPixmap(get_image_file(image_filename))



from typing import Any, Self

from PyQt6.QtWidgets import QFormLayout

from .....classes import DescriptionInfo
from .....custom_widgets import SpinBox
from .....utility import layout as layout_util
from ...base_widget import AbstractBaseWidget
from . import encoding
from .process import HoldProcess


class HoldWidget(AbstractBaseWidget):
    """Hold for a duration."""

    BASE_DISPLAY_NAME = "Hold"

    def __init__(self):
        layout = QFormLayout()
        AbstractBaseWidget.__init__(
            self,
            layout,
            self.BASE_DISPLAY_NAME,
            HoldProcess,
            "clock-select.png",
            DescriptionInfo(
                "flow_control",
                "hold",
                HoldProcess.directory_name(),
                DescriptionInfo.Substitutions(
                    parameters_dict={
                        "HOURS": encoding.HOURS,
                        "MINUTES": encoding.MINUTES,
                        "SECONDS": encoding.SECONDS,
                    }
                ),
            ),
        )

        self.hours_spinbox = SpinBox()
        self.minutes_spinbox = SpinBox()
        self.seconds_spinbox = SpinBox()
        self.spinboxes = (self.hours_spinbox, self.minutes_spinbox, self.seconds_spinbox)
        layout_util.add_to_form_layout(
            layout,
            (encoding.HOURS, self.hours_spinbox),
            (encoding.MINUTES, self.minutes_spinbox),
            (encoding.SECONDS, self.seconds_spinbox),
        )

        for spinbox in self.spinboxes:
            spinbox.textChanged.connect(self.handle_value_change)

    def handle_value_change(self, *args):
        """Handle any of the duration spinboxes changing."""
        hours = self.hours_spinbox.text()
        minutes = self.minutes_spinbox.text()
        seconds = self.seconds_spinbox.text()
        self.setWindowTitle(
            f"{self.BASE_DISPLAY_NAME} ({hours} hours, {minutes} minutes, {seconds} seconds)"
        )

    @classmethod
    def from_dict(cls, data_as_dict: dict[str, Any]) -> Self:
        widget = cls()
        widget.hours_spinbox.setValue(data_as_dict[encoding.HOURS])
        widget.minutes_spinbox.setValue(data_as_dict[encoding.MINUTES])
        widget.seconds_spinbox.setValue(data_as_dict[encoding.SECONDS])
        return widget

    def to_dict(self) -> dict[str, Any]:
        return {
            encoding.HOURS: self.hours_spinbox.value(),
            encoding.MINUTES: self.minutes_spinbox.value(),
            encoding.SECONDS: self.seconds_spinbox.value(),
        }



from PyQt6.QtWidgets import QGroupBox, QLayout, QSizePolicy


class GroupBox(QGroupBox):
    """
    `QGroupBox` with that automatically sets the size policy and title.

    Parameters
    ----------
    title
        The window's title.
    layout
        The layout to initialize with.
    """

    def __init__(self, title: str | None, layout: QLayout):
        QGroupBox.__init__(self)
        self.setTitle(title)
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.setLayout(layout)
